#!/usr/bin/env python
# coding: utf-8

# In[1]:


import os

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Rectangle

import functools

import genx_fusion_analysis_base as gnx
from genx_fusion_analysis_base import colors


# In[2]:


colors = gnx.colors

nice_gray = "#333333"

mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.titlesize'] = 14


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = 14
mpl.rcParams['ytick.labelsize'] = 14
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 13


# In[3]:


x_all = np.r_[10,20,50,100,150,200,250,300,350]


# In[4]:


a_all = np.arange(1, 10)


# In[5]:


scen_med = "MedOpp"


# # Series 10000 build-outs

# In[6]:


base_net_mwe = 100e3
def generator_per_netmwe_10k(scen, cases):
    gen = []
    for i in cases:
        if i >= 10000 and i < 20000 or i==4 or i==104:
            divisor = 100e3
        elif i>=20000 and i < 20100:
            divisor = 20e3
        elif i>=20100 and i < 20200:
            divisor = 50e3
        elif i>=20200 and i < 20300:
            divisor = 100e3
        elif i>=20300 and i < 20400:
            divisor = 150e3
        elif i>=20400 and i < 20500:
            divisor = 200e3
        else:
            raise Exception("Error, not programmed yet")
        gen.append(gnx.grab_data('generation', scen, [i])[0] / divisor)
    return np.array(gen)

def additional_generation_cost(scen, cases):
    return gnx.base_gen_cost * generator_per_netmwe_10k(scen, cases)


# In[7]:


s10k = 10000
midstor_buildout = 20
nostor = 100
midstor = 200


# In[8]:


mid_vom = 3
mid_pulse_length = 4
mid_dwell = 0.15
mid_pass_recirc = 0.1
mid_act_recirc = 0.1
mid_start_energy = 0.025


# # Series 10100: mid-range fusion without storage

# In[9]:


output_data = dict()


# In[10]:


my_scen = scen_med

core_param_df = gnx.load_core_parameters_df(my_scen)


# ### Core min power

# In[11]:


series10100_cases = s10k + nostor + np.array(range(0,6))
series10100_coremin_power = [0.2,0.4,0.6,0.8,0.9,1.0]
series10100_duals = gnx.grab_duals(scen_med, series10100_cases)


# ### Pulse duration

# In[12]:


def int_list(a):
    return [int(el) for el in a]

def float_list(a):
    return [float(el) for el in a]


# In[13]:


nostor_reference = 10004
midstor_reference = 10024


# In[14]:


series10120_cases = np.insert(s10k + nostor + 20 + np.arange(0,6), 3, nostor_reference)
series10120_cases = np.insert(series10120_cases, 6, 10126)
series10120_durations = [1,2,3,4,6,12,24,-1]

series10120_duals = gnx.grab_duals(my_scen, series10120_cases)

series10120_vom = [gnx.totalVOMcost(mid_vom, gnx.base_η, pd, mid_dwell, mid_pass_recirc, mid_act_recirc, mid_start_energy) for pd in series10120_durations]
series10120_totcost = series10120_duals + additional_generation_cost(my_scen, series10120_cases)


# In[15]:


output_data["Pulse cycle duration"] = {'cases': int_list(series10120_cases),
                               'durations': int_list(series10120_durations),
                               'x': float_list(series10120_vom),
                               'y': float_list(series10120_totcost)}


# ### CoreDwellTime

# In[16]:


series10130_cases = np.insert(s10k + nostor + 30 + np.arange(0,6), 2, nostor_reference)
series10130_dwelltimes = [0.,0.1,0.15,0.2,0.3,0.4,0.6]
series10130_duals = gnx.grab_duals(scen_med, series10130_cases)
series10130_vom = [gnx.totalVOMcost(mid_vom, gnx.base_η, mid_pulse_length, dw, mid_pass_recirc, mid_act_recirc, mid_start_energy) for dw in series10130_dwelltimes]
series10130_totcost = series10130_duals + additional_generation_cost(my_scen, series10130_cases)


# In[17]:


output_data["Dwell time"] = {'cases': int_list(series10130_cases),
                               'dwell_time': float_list(series10130_dwelltimes),
                               'x': float_list(series10130_vom),
                               'y': float_list(series10130_totcost)}


# ### CoreRecircPass

# In[18]:


series10140_cases = np.insert(s10k + nostor + 40 + np.arange(0,5), 2, nostor_reference)
series10140_recircpass = [0.,0.05,0.1,0.15,0.2,0.3]
series10140_duals = gnx.grab_duals(scen_med, series10140_cases)
series10140_vom = [gnx.totalVOMcost(mid_vom, gnx.base_η, mid_pulse_length, mid_dwell, x, mid_act_recirc, mid_start_energy) for x in series10140_recircpass]
series10140_totcost = series10140_duals + additional_generation_cost(my_scen, series10140_cases)


# In[19]:


series10140_cases


# In[20]:


output_data["Passive recirculating power"] = {'cases': int_list(series10140_cases),
                               'recirc_pass': float_list(series10140_recircpass),
                               'x': float_list(series10140_vom),
                               'y': float_list(series10140_totcost)}


# ### CoreRecircAct

# In[21]:


series10150_cases = np.insert(s10k + nostor + 50 + np.arange(0,5), 2, nostor_reference)
series10150_recircact = [0.,0.05,0.1,0.15,0.2,0.3]
series10150_duals = gnx.grab_duals(my_scen, series10150_cases)
series10150_vom = [gnx.totalVOMcost(mid_vom, gnx.base_η, mid_pulse_length, mid_dwell, mid_pass_recirc, x, mid_start_energy) for x in series10150_recircact]
series10150_totcost = series10150_duals + additional_generation_cost(my_scen, series10150_cases)


# In[22]:


series10150_cases


# In[23]:


output_data["Active recirculating power"] = {'cases': int_list(series10150_cases),
                               'recirc_act': series10150_recircact,
                               'x': float_list(series10150_vom),
                               'y': float_list(series10150_totcost)}


# ### CoreStartPower

# In[24]:


series10160_cases = np.insert(s10k + nostor + 60 + np.arange(0,5), 2, nostor_reference)
series10160_corestartpower = [0.,0.05,0.1,0.2,0.3,0.4]
series10160_duals = gnx.grab_duals(scen_med, series10160_cases)
series10160_vom = [gnx.totalVOMcost(mid_vom, gnx.base_η, mid_pulse_length, mid_dwell, mid_pass_recirc, mid_act_recirc, mid_start_energy) for _ in series10160_corestartpower]
series10160_totcost = series10160_duals + additional_generation_cost(my_scen, series10160_cases)


# In[25]:


output_data["Core start power"] = {'cases': int_list(series10160_cases),
                               'start_power': series10160_corestartpower,
                               'x': float_list(series10160_vom),
                               'y': float_list(series10160_totcost)}


# ### CoreStartEnergy

# In[26]:


series10170_cases = np.insert(s10k + nostor + 70 + np.arange(0,4), 1, nostor_reference)
series10170_corestartenergy = [0.,0.025,0.1,0.15,0.2]
series10170_duals = gnx.grab_duals(scen_med, series10170_cases)
#series10170_corevom = np.nan#[effectiveVOMcost(base_vom, base_η, base_pulse_length, base_dwell, base_pass_recirc, base_act_recirc, x) for x in series10170_corestartenergy]

series10170_vom = [gnx.totalVOMcost(mid_vom, gnx.base_η, mid_pulse_length, mid_dwell, mid_pass_recirc, mid_act_recirc, x) for x in series10170_corestartenergy]
series10170_totcost = series10170_duals + additional_generation_cost(my_scen, series10170_cases)


# In[27]:


output_data["Core start energy"] = {'cases': int_list(series10170_cases),
                                   'start_energy': series10170_corestartenergy,
                                   'x': float_list(series10170_vom),
                                   'y': float_list(series10170_totcost)}


# ### Core VarOMCostPerMWhth

# In[28]:


get_core_vom = functools.partial(gnx.core_vom_cost, core_param_df)

series10180_cases = [10180, 10181, 10182, nostor_reference, 10184, 10185, 10186, 10187]
series10180_vomcostpermwhth = list(map(get_core_vom, series10180_cases))
series10180_duals = gnx.grab_duals(scen_med, series10180_cases)
series10180_vom = [gnx.totalVOMcost(x, gnx.base_η, mid_pulse_length, mid_dwell, mid_pass_recirc, mid_act_recirc, mid_start_energy) for x in series10180_vomcostpermwhth]
series10180_totcost = series10180_duals + additional_generation_cost(my_scen, series10180_cases)


# In[29]:


output_data["Core VOM cost"] = {'cases': int_list(series10180_cases),
                               'vom_cost': float_list(series10180_vomcostpermwhth),
                               'x': float_list(series10180_vom),
                               'y': float_list(series10180_totcost)}


# ### Idealized plant

# In[30]:


series19180_cases = list(range(19180, 19186))

series19180_vomcostpermwhth = list(map(get_core_vom, series19180_cases))

series19180_duals = gnx.grab_duals(scen_med, series19180_cases)


# In[31]:


series19180_vom = [gnx.totalVOMcost(x, gnx.base_η, -1, 0, 0, 0, 0) for x in series19180_vomcostpermwhth]
series19180_totcost = series19180_duals + additional_generation_cost(my_scen, series19180_cases)
output_data["Idealized plant"] = {'cases': int_list(series19180_cases),
                               'core_vom_cost': float_list(series19180_vomcostpermwhth),
                               'x': float_list(series19180_vom),
                               'y': float_list(series19180_totcost)}


# # Text output

# In[32]:


chart_name = "3"


# In[33]:


header = f"""Figure {chart_name}:

Value of the mid-range plant with some variations in its core parameter set.
""" + gnx.generic_data_header + \
f"""Each of the seven dicts provides the data for one of the varied parameters.
Each set contains the case numbers (see Step3/MedOpp/replacements.csv), the varied parameters,
and the x and y values for the respective points.
x values are in $/MWh and y values are in $/MW·yr.

Subpart (a) contains also the points for the standard reference reactors: 
see data from Figure 1 at the point of 100GW of capacity penetration, in the MedOpp scenario.
"""

def preprend_hash(string):
    lines = string.split('\n')
    commented_lines = ['# ' + l for l in lines]
    return "\n".join(commented_lines)

import yaml
output = yaml.dump(output_data)

with open(gnx.product_path(f"Figure_{chart_name}.yml"), 'w') as f:
    f.write(preprend_hash(header) + '\n')
    f.write(output)


# # Comparison with existing resources in the mid case

# The NY_E_utilitypv_class1_moderate_90_3 has 2.5x the cost of other options. Not sure why.

# In[34]:


def tableaux_positions(scen, cases):
    core_parameters = gnx.load_core_parameters_df(scen)
    vomcostpermwhth = [gnx.core_vom_cost(core_parameters, kase) for kase in cases]
    
    thermal_eff = [gnx.lookup_by_column(core_parameters, kase, "ThermalEff") for kase in cases]
    core_max_up = [gnx.lookup_by_column(core_parameters, kase, "CoreMaxUp") for kase in cases]
    dwell_time = [gnx.lookup_by_column(core_parameters, kase, "CoreDwellTime") for kase in cases]
    pass_recirc = [gnx.get_passive_recirc_fraction(core_parameters, kase) for kase in cases]
    act_recirc = [gnx.lookup_by_column(core_parameters, kase, "CoreRecircAct") for kase in cases]
    core_start_energy = [gnx.lookup_by_column(core_parameters, kase, "CoreStartEnergy") for kase in cases]   
    
    duals = gnx.grab_duals(scen, cases)
    vom = [gnx.totalVOMcost(*_) for _ in zip(vomcostpermwhth, thermal_eff, core_max_up, dwell_time, pass_recirc, act_recirc, core_start_energy)]

    totcost = duals + additional_generation_cost(scen, cases)
    return vom, totcost/1e3


# In[35]:


cases_100GW = [20200, 20201, 20202, 20203, 20204, 20205]


# In[36]:


def plot_variation_line(ax, scen, cases, **kwargs):  
    x, y = tableaux_positions(scen, cases)
    ax.plot(x, y, **kwargs)


# In[37]:


def plot_variation_lines(ax, scen):
        
    def plot_variation_line(ax, cases, label=None, **kwargs):  
        x, y = tableaux_positions(scen, cases)
        ax.plot(x, y, **kwargs)
                   
    plot_variation_line(ax, cases_100GW, '100 GW', color=colors[1])   


# # Figure 3

# In[38]:


my_scen = "MedOpp"
myxlabel = "Marginal cost of net generation, $/MWh"
myylabel = "Plant value (or cost threshold), $/MW·yr"


# In[63]:


fig = plt.figure(figsize=(9.5, 5.))

plt.rcParams['text.usetex']=False


gs = GridSpec(2, 3, width_ratios=[1.8,1,1], height_ratios=[0.8, 1], wspace=0.25)

ax1 = fig.add_subplot(gs[:, 0])
ax2 = fig.add_subplot(gs[0,1])
ax3 = fig.add_subplot(gs[0,2])

#plot_variation_lines(ax1, my_scen)

point_reactor_size = 80

opt_marker = 'P'
mid_marker = '*'
pes_marker = 's'

ax1.scatter([4.43], (gnx.grab_duals(my_scen, [104])+additional_generation_cost(my_scen, [104]))/1e3,
           point_reactor_size, marker=opt_marker,)
ax1.scatter([series10180_vom[3]], [(series10180_totcost[3])/1e3],
           point_reactor_size, marker=mid_marker,)
[ax.scatter([series10180_vom[3]], [(series10180_totcost[3])/1e3], 
           point_reactor_size*0.7, marker=mid_marker, color=colors[1], zorder=100) for ax in [ax2, ax3]]

ax1.scatter([25.58], (gnx.grab_duals(my_scen, [4])+additional_generation_cost(my_scen, [4]))/1e3,
           point_reactor_size, marker=pes_marker,)

#ax1.plot(series10180_vom[2:5], series10180_totcost[2:5]/1e3, color=colors[1], zorder=-10, label="Test")

ax1.plot(series19180_vom, series19180_totcost/1e3, color='gray', zorder=-10)


ax1.set_ylabel(myylabel)

ax1.set_xlim(0, 50)
ax1.set_ylim(0,500)

#x1, x2, y1, y2 = 8.3,16.5,370, 430
x1, x2, y1, y2 = 8.0,18.3,359, 439

ax2.set_xlim(x1, x2)
ax2.set_ylim(y1, y2)
r1 = Rectangle((x1, y1), x2-x1, y2-y1, fc='white', fill=False, edgecolor='gray', zorder=-2)
ax1.add_artist(r1)

labelsize = 12
ax1.text(x1-1, y1-10, 'See (b)', ha='left', va='top', color='gray', style='italic', size=labelsize)


x1, x2, y1, y2 = 11.4,12.7, 398, 414.5
ax3.set_xlim(x1, x2)
ax3.set_ylim(y1, y2)
r2 = Rectangle((x1, y1), x2-x1, y2-y1, fc='white', fill=False, edgecolor='gray', zorder=-2)
ax2.add_artist(r2)

ax2.text((x1+x2)/2 - 0.5, y1-2, 'See (c)', ha='center', va='top', color='gray', style='italic', size=labelsize)


def plot_endpoints(ax, x, y, markersize=30, **kwargs):
    xi, *_x, xf = x
    yi, *_y, yf = y
    ax.plot(x, y, lw=0.3, color='#888888')
    
    del kwargs['color']
    ax.scatter([xi, xf], [yi, yf], markersize, color=nice_gray, **kwargs)

sl = slice(2,5)
ax2.plot(series10180_vom[sl], series10180_totcost[sl]/1e3, color=colors[1], label="Core VOM", zorder=-100)
ax3.plot(series10180_vom[sl], series10180_totcost[sl]/1e3, color=colors[1])
plot_endpoints(ax2, series10140_vom, series10140_totcost/1e3, color=colors[3], label="Passive recirc")
plot_endpoints(ax2, series10150_vom, series10150_totcost/1e3, color=colors[4], label="Active recirc", marker='s', facecolor='white')
plot_endpoints(ax3, series10120_vom[:-2], series10120_totcost[:-2]/1e3, color=colors[5], label="Pulse length", marker='^')
plot_endpoints(ax3, series10130_vom, series10130_totcost/1e3, color=colors[6], label="Dwell time", marker='p', facecolor='white', markersize=40)
plot_endpoints(ax3, series10170_vom, series10170_totcost/1e3, color=colors[7], label="Core start energy", marker='d')
plot_endpoints(ax3, series10160_vom, series10160_totcost/1e3, color=colors[8], label="Core start power", marker='x')

fig.supxlabel(myxlabel, fontsize=14)

ax1.text(0.97, 0.97, '(a)', ha='right', va='top', transform=ax1.transAxes)

f_ms = 8
legend_elements = [Line2D([0], [0], marker=opt_marker, ms=f_ms, color=colors[0], lw=0, label='Optimistic'),
                   Line2D([0], [0], marker=mid_marker, ms=f_ms, color=colors[1], lw=0, label='Mid-range'),
                   Line2D([0], [0], marker=pes_marker, ms=f_ms ,color=colors[2], lw=0, label='Pessimistic'),]

ax1.legend(handles=legend_elements,
           loc=(0.52, 0.68),
           ncol=1, fontsize=12,
           frameon=True,
           columnspacing=0.6,
           handlelength=0.4,
           labelspacing=0.5)

ax2.text(0.97, 0.97, '(b)', ha='right', va='top', transform=ax2.transAxes)
ax3.text(0.97, 0.97, '(c)', ha='right', va='top', transform=ax3.transAxes)

#ax1.text(0.47, 0.36, '100 GW fusion cap.', ha='left', va='top', transform=ax1.transAxes,color='gray', style='italic', )

ax1.text(0.67, 0.5, 'Idealized\nplants', ha='left', va='top', transform=ax1.transAxes,color='gray', style='italic', )


#ax1.text(0.73, 0.37, 'Mid-range\nplant with\naltered $\pi^{VOM, th}$', ha='center', va='top', transform=ax1.transAxes,color=colors[1], style='italic', )


mpl.rcParams['text.usetex']=True
mpl.rc('text.latex', preamble=r"""\usepackage{siunitx}"""     r'\usepackage{dcolumn}'     r'\newcolumntype{d}[1]{D{.}{.}{#1}}'     r'\usepackage{stix}' )
       

s1 = r"""\begin{tabular}{@{}c l r r r@{}} """ r""" & Core parameter &  \multicolumn{1}{c}{Best} &  Ref.\ (\ \ )  & \multicolumn{1}{r@{}}{Worst} \\ \hline """ r""" & $\pi^{VOM, th}\;\cdot$ MWh$_{(\mathrm{th})}$ & \$2 & \$3 & \$5 \\ """ r"""$\mdlgblkcircle$ & Pass. recirc frac.& 0 & 0.1 &  0.3 \\ """ r"""$\mdwhtsquare$ & Act. recirc frac.& 0 & 0.1 &  0.3 \\ """ r"""$\blacktriangle$ & Pulse length / h &  12 & 4 & 1 \\ """ r"""$\pentagon$ & Dwell time / h &  0 & 0.15 & 0.6 \\ """ r"""$\mdblklozenge$ & Start energy frac. &  0 &  0.025 & 0.2 \\ """ r"""$\times$ & Start power frac. &  0 & 0.1 & 0.4 \\ """ r"""\end{tabular}"""
y_adj = -0.025

ax1.text(1.2,y_adj, s1, ha="left",va="bottom",transform=ax1.transAxes, size=14, color=nice_gray)

yline = y_adj + 0.40
center = 1 + 0.225
width = 0.05
ax1.plot([center-width/2,center+width/2], 2*[yline], transform=ax1.transAxes, color=colors[1], clip_on=False)
ax1.scatter([2.205], [y_adj + 0.478], 80, transform=ax1.transAxes, color=colors[1], clip_on=False, marker="*")

chart_name = "3"

plt.savefig(gnx.product_path(f'Figure_{chart_name}.pdf'), bbox_inches='tight')

plt.show()


# In[53]:


def plot_variation_lines(ax, scen):
        
    def plot_variation_line(ax, cases, label=None, **kwargs):  
        x, y = tableaux_positions(scen, cases)
        y = y / 1e3 / gnx.CAPEX_TO_ANN
        ax.plot(x, y, **kwargs)
                   
    plot_variation_line(ax, cases_100GW, '100 GW', color=colors[1])   


# In[146]:


myxlabel = r"Variable cost, $/MWh"
myylabel = r"Capital cost threshold, $/W"


# In[149]:


mpl.rcParams['text.usetex']=False

fig, ax = plt.subplots(figsize=(5, 5.))

plot_variation_lines(ax, my_scen)

point_reactor_size = 80

opt_marker = 'P'
mid_marker = '*'
pes_marker = 's'

ax.scatter([4.43], (gnx.grab_duals(my_scen, [104])+additional_generation_cost(my_scen, [104]))/1e6/0.078,
           point_reactor_size, marker=opt_marker,)
ax.scatter([series10180_vom[3]], [(series10180_totcost[3])/1e6/0.078],
           point_reactor_size, marker=mid_marker,)
ax.scatter([25.58], (gnx.grab_duals(my_scen, [4])+additional_generation_cost(my_scen, [4]))/1e6/0.078,
           point_reactor_size, marker=pes_marker,)


ax.set_ylabel(myylabel)

ax.set_xlim(0, 50)
ax.set_ylim(0,6.5)

fig.supxlabel(myxlabel)

f_ms = 8
legend_elements = [Line2D([0], [0], marker=opt_marker, ms=f_ms, color=colors[0], lw=0, label='Optimistic'),
                   Line2D([0], [0], marker=mid_marker, ms=f_ms, color=colors[1], lw=0, label='Mid-range'),
                   Line2D([0], [0], marker=pes_marker, ms=f_ms ,color=colors[2], lw=0, label='Pessimistic'),]

ax.legend(handles=legend_elements,
           loc=(0.52, 0.68),
           ncol=1, fontsize=12,
           frameon=True,
           columnspacing=0.6,
           handlelength=0.4,
           labelspacing=0.5)

ax.text(0.73, 0.40, 'Mid-range\nplant with\naltered variable cost',
        ha='center', va='top',
        transform=ax.transAxes,color='gray', style='italic', )

plt.savefig(gnx.product_path('Figure_3_simple.pdf'), bbox_inches='tight')

plt.show()

