#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx


# In[ ]:


import numpy as np
import matplotlib.pyplot as plt

from matplotlib.lines import Line2D
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Patch
from matplotlib.patches import Rectangle

import matplotlib.patches as mpatches

import matplotlib as mpl

from copy import deepcopy

import colorsys
import functools
import os

import pandas as pd


# Note: these cases have 6 hour ... 1 hour up/down times.

# In[ ]:


colors = gnx.colors

nice_gray = "#333333"

#mpl.rcParams['font.family'] = 'sans-serif'
mpl.rcParams['font.size'] = 8
mpl.rcParams['axes.titlesize'] = 10


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = 8
mpl.rcParams['ytick.labelsize'] = 8
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 8


# In[ ]:


def generator_per_netmwe_10k(scen, indices):
    gen = []
    for i in indices:
        if i >= 10000 and i < 20000 or i==4 or i==104:
            divisor = 100e3
        elif i>=20000 and i < 20100:
            divisor = 20e3
        elif i>=20100 and i < 20200:
            divisor = 50e3
        elif i>=20200 and i < 20300:
            divisor = 100e3
        elif i>=20300 and i < 20400:
            divisor = 150e3
        elif i>=20400 and i < 20500:
            divisor = 200e3
        elif i>=20900 and i < 20906:
            divisor = 10e3
        else:
            raise Exception("Error, not programmed yet")
        gen.append(gnx.grab_data('generation', scen, [i])[0] / divisor)
    return np.array(gen)

def additional_generation_cost(scen, indices):
    return gnx.base_gen_cost * generator_per_netmwe_10k(scen, indices)


# In[ ]:


η = 0.4

mid_vom = 3
mid_pulse_length = 4
mid_dwell = 0.15
mid_pass_recirc = 0.1
mid_act_recirc = 0.1
mid_start_energy = 0.025

def tableaux_positions(scen, cases):
    cases_vomcostpermwhth = [0.5, 2, 3, 5, 8, 13]
    cases_duals = gnx.grab_duals(scen, cases)
    cases_vom = [gnx.totalVOMcost(x, gnx.base_η, mid_pulse_length, mid_dwell, mid_pass_recirc, mid_act_recirc, mid_start_energy) for x in cases_vomcostpermwhth]
    cases_totcost = cases_duals + additional_generation_cost(scen, cases)
    return cases_vom, cases_totcost/1e3
    


# # Comparison with existing resources in the mid case

# In[ ]:


@functools.cache
def nonfusion_new_build_generators_data(scen):
    gen_dat = gnx.load_generators_data(scen)
    new_builds = gen_dat[gen_dat.New_Build > 0]
    num_zones = 20
    return new_builds.head(len(new_builds) - num_zones)

def cost_per_MMBTU(fuels_df, fuel):
    r"""Extract yearly mean price of fuel from fuels dataframe
    """
    return np.mean(fuels_df[fuel][1:])

#------------------

def string_in_array(s, a):
    res = []
    for el in a:
        res.append(s in el)
    return np.array(res)

def get_rows_for_resource(gen_df, resource):
    """Generators_data view for a resource type
    
    Parameters
    ----------
    gen_df : pd.DataFrame
        Geneartors dataframe
    
    resource: string
        like frcc_biomass_1
    """
    sel = string_in_array(resource, gen_df.Resource.to_numpy())
    return gen_df[sel]

def annual_cost_for_resource_per_mwyr(gen_df, resource):
    r = get_rows_for_resource(gen_df, resource)
    inv = r.Inv_Cost_per_MWyr
    fom = r.Fixed_OM_Cost_per_MWyr
    ann = np.array(inv, dtype=float) + np.array(fom, dtype=float)
    capfactor = np.array(r.MeanCapFactor, dtype=float)
    return ann / capfactor

@functools.cache
def vom_for_resource_per_mwh(scen, resource):
    r"""
    Parameters
    ----------
    scen : string
        Scenario folder
    resource: string
        like frcc_biomass_1
    """
    gen_df = nonfusion_new_build_generators_data(scen)
    
    r = get_rows_for_resource(gen_df, resource)
    varom = np.array(r.Var_OM_Cost_per_MWh, dtype=float)
    heat_rate = r.Heat_Rate_MMBTU_per_MWh
    heat_rate = np.array(heat_rate, dtype=float)
    fuels = r.Fuel
    
    fuels_df = gnx.load_fuels_data(scen)
    
    mmbtu_cost = np.array([cost_per_MMBTU(fuels_df, f) for f in fuels], dtype=float)
    return varom + heat_rate * mmbtu_cost

def plot_var_ann(ax, scen, resource, markersize=8, **kwargs):
    x = vom_for_resource_per_mwh(scen, resource)
    gen_df = nonfusion_new_build_generators_data(scen)
    y = annual_cost_for_resource_per_mwyr(gen_df, resource)
    
    lab = gnx.categories_table[resource]
    color = gnx.color_map[lab]
    
    ax.scatter(x, y/1e3, markersize, color=color, **kwargs)


# In[ ]:


def scen_resource_rows(scen, res):
    x = np.round(vom_for_resource_per_mwh(scen, res), 3)
    gen_df = nonfusion_new_build_generators_data(scen)
    y = annual_cost_for_resource_per_mwyr(gen_df, res)
    return [(scen, res, xx, int(yy)) for xx, yy in zip(x, y)]


# In[ ]:


def scale_lightness(colorname, scale_l):
    rgb = mpl.colors.ColorConverter.to_rgb(colorname)
    # convert rgb to hls
    h, l, s = colorsys.rgb_to_hls(*rgb)
    # manipulate h, l, s values and return as rgb
    return colorsys.hls_to_rgb(h, min(1, l * scale_l), s = s)

fusion_colors = [scale_lightness(gnx.color_map[gnx.FUSION], i) for i in [1.0, 0.85, 0.7]]


# In[ ]:


cases_10_gw = [20900, 20901, 20902, 20903, 20904, 20905]
cases_100_gw = [20200, 20201, 20202, 20203, 20204, 20205]
cases_200_gw = [20400, 20401, 20402, 20403, 20404, 20405]


# In[ ]:


def plot_variation_lines(ax, scen, annotate=False):
    x_shift = 5
    y_shift = -20
    
    def annotate_last_y(x, y, label):
        ax.text(x[-1]+x_shift, y[-1]+y_shift, label)
        
    def plot_variation_line(ax, cases, label=None, **kwargs):  
        x, y = tableaux_positions(scen, cases)
        ax.plot(x, y, **kwargs)
        if annotate and label is not None:
            annotate_last_y(x, y, label)
    
    plot_variation_line(ax, cases_10_gw, '10 GW', color=fusion_colors[0], lw=0.75)
    plot_variation_line(ax, cases_100_gw, '100 GW', color=fusion_colors[1], lw=0.75)
    plot_variation_line(ax, cases_200_gw, '200 GW', color=fusion_colors[2], lw=0.75)   


# In[ ]:


def plot_competitor_points_and_fusion_lines(ax, scen, annotate_fusion=False):
    
    s_s= 10
    w_s = 25
    n_s = 10
    gas_size = 10
    
    solar_marker = 'o'
    wind_marker = '2'    
    
    plot_var_ann(ax, scen, "landbasedwind_class1_advanced", markersize=w_s, marker=wind_marker, linewidths=1) 
    plot_var_ann(ax, scen, "offshorewind_class1_advanced_fixed", markersize=w_s, marker=wind_marker, linewidths=1)     
    
    plot_var_ann(ax, scen, "landbasedwind_class1_moderate", markersize=w_s, marker=wind_marker, linewidths=1)    
    plot_var_ann(ax, scen, "offshorewind_class1_moderate_fixed", markersize=w_s, marker=wind_marker, linewidths=1)     

    plot_var_ann(ax, scen, "utilitypv_class1_advanced", markersize=s_s, marker=solar_marker, linewidths=0)   
    plot_var_ann(ax, scen, "utilitypv_class1_moderate", markersize=s_s, marker=solar_marker, linewidths=0)   


    plot_var_ann(ax, scen, "nuclear_nuclear_moderate", marker='D', markersize=n_s, linewidths=0)   
    plot_var_ann(ax, scen, "naturalgas_ccs100_moderate", marker=">", markersize=12, linewidths=0)
    plot_var_ann(ax, scen, "naturalgas_ctavgcf_moderate", marker="+", markersize=1.2*gas_size, lw=0.5)
    plot_var_ann(ax, scen, "naturalgas_ccavgcf_moderate", marker="x", markersize=gas_size, lw=0.5)   
    plot_variation_lines(ax, scen, annotate=annotate_fusion)


# In[ ]:


deflation_2020_to_2019 = 255.658 / 258.812  

def plot_aries_at_point(ax):
    ARIES_AT_point = (8.8256748789 * deflation_2020_to_2019, 418.93278472019995 * deflation_2020_to_2019)
    kw = {'ha': 'center', 'va': 'top', 'size': 10, 'color':'black'}
    l,  = ax.plot([ARIES_AT_point[0]], [ARIES_AT_point[1]], marker='o',
                  linewidth=0.5,
                  markerfacecolor='white',markeredgewidth=0.5,
                  markeredgecolor='#222222',
                  zorder=12)
    l.set_markerfacecolor((1, 1, 1, 0.0))    


# In[ ]:


myxlabel = "Variable costs, $/MWh"
myylabel = "Plant cost (or value), $/MW·yr"


# In[ ]:


def num_to_lowercase(n):
    return chr(ord("`") + 1 + n)


# # Figure scenario_variations

# In[ ]:


fig, axs = plt.subplots(3, 1, figsize=(3.75, 5), sharey=True, sharex=True, squeeze=False)

plot_competitor_points_and_fusion_lines(axs[0,0], "LowOpp", annotate_fusion=False)
plot_competitor_points_and_fusion_lines(axs[1,0], "MedOpp")
plot_competitor_points_and_fusion_lines(axs[2,0], "HighOpp")

min_x = -2
max_x = 200
axs[0,0].set_xlim(min_x, max_x)
axs[1,0].set_xlim(min_x, max_x)
axs[2,0].set_xlim(min_x, max_x)
axs[1,0].set_ylim(0,720)

labelsize = 12



[ax.spines["top"].set_visible(False) for ax in axs.reshape(np.size(axs))]
[ax.spines["right"].set_visible(False) for ax in axs.reshape(np.size(axs))]
#[ax.spines["left"].set_visible(False) for ax in axs.reshape(np.size(axs))]
[ax.spines["bottom"].set(bounds=[0,max_x]) for ax in axs.reshape(np.size(axs))]

for ax in axs.reshape(np.size(axs)):
    ds = [10, 4]
    plot_aries_at_point(ax)
#     for coe in [50, 100,150,200,250]:
#         gnx.plot_lcoe_line(ax, coe, color='lightgray', lw=0.5, capacityfactor=1)


    
axbs = axs.reshape(np.size(axs))
scen_group_labels = ["Low market opp.", "Medium market opp.", "High market opp."]

for i, ax in enumerate(axbs):
    ax.text(0.125, 0.88, "("+num_to_lowercase(i)+") " + scen_group_labels[i], ha='left', va='bottom', transform=ax.transAxes)

axs[2,0].set(xlabel=myxlabel)
fig.supylabel(myylabel, x=0.00, fontsize=8)

my_markers=["2", "o", "D", ">", "x", "+"]
my_markersizes = [6, 4,  3, 3, 4, 5]

point_reactors = [gnx.WIND, gnx.SOLAR, gnx.NUCLEAR, gnx.CCS, gnx.ZCFCC, gnx.ZCFCT]
leg_points = [Line2D([0], [0], marker=my_markers[i],
                     color=gnx.color_map[res], lw=0,
                     label=res,
                     markerfacecolor=gnx.color_map[res],
                     markersize=my_markersizes[i]) for i, res in enumerate(point_reactors)]

legend_elements = [Line2D([0], [0], color=fusion_colors[0], lw=0.75, label='Mid-range-like\nfusion, 10 GW'),
                   Line2D([0], [0], color=fusion_colors[1], lw=0.75, label='100 GW'),
                   Line2D([0], [0], color=fusion_colors[2], lw=0.75, label='200 GW'),
                   ]

legend_elements = legend_elements + leg_points

axs[1,0].legend(handles=legend_elements,loc=(0.6, 0.55), framealpha=1)

axs[0,0].annotate("ARIES-AT", (8, 420), color="#222222", xytext=(12, 430))

product_name = "4"

plt.savefig(gnx.product_path(f"Figure_{product_name}.pdf"), bbox_inches='tight')

plt.show()


# In[ ]:


cases = {"10GW": cases_10_gw, "100GW": cases_100_gw, "200GW": cases_200_gw}

scenarios_to_load = [
     "LowOpp",
     "MedOpp",
     "HighOpp",
 ]

data = {scen + "_" + pen: tableaux_positions(scen, case_nums)[1]
        for scen in scenarios_to_load
        for pen, case_nums in cases.items()}

data["VariableCosts"] = tableaux_positions("MedOpp", cases_10_gw)[0]

df = pd.DataFrame(data)
a = df.to_numpy()
columns = ", ".join(df.keys())

header = f"""Figure {product_name}:
Fusion value and competitor costs.
""" + gnx.generic_data_header + \
f"""

""" + \
columns

np.savetxt(gnx.product_path(f"Figure_{product_name}_lines.csv"), a, fmt="%.1f", header=header, delimiter=',')


# In[ ]:


from itertools import chain


# In[ ]:


resources_to_list = ("landbasedwind_class1_advanced",
    "offshorewind_class1_advanced_fixed",
    "landbasedwind_class1_moderate",
    "offshorewind_class1_moderate_fixed", 
    "utilitypv_class1_advanced",
    "utilitypv_class1_moderate", 
    "nuclear_nuclear_moderate",
    "naturalgas_ccs100_moderate",
    "naturalgas_ctavgcf_moderate",
    "naturalgas_ccavgcf_moderate",
)

big_list = [scen_resource_rows(scen, res) for scen in scenarios_to_load for res in resources_to_list]

no_empty_lists = [a for a in big_list if len(a) > 0]
flattened = np.array(list(chain(*no_empty_lists)))

columns = "Scenario, Resource, Variable costs / ($/MWh), Annual cost / ($/MW·yr)"

header = f"""Figure {product_name}:
Fusion value and competitor costs.
""" + gnx.generic_data_header + \
f"""

""" + \
columns

np.savetxt(gnx.product_path(f"Figure_{product_name}_points.csv"), flattened,
           fmt="%s", delimiter=',', header=header)


# # Plot for supplementary data

# In[ ]:


chart_name = "SUPP_competition"


# In[ ]:


fig, axs = plt.subplots(4, 3, figsize=(7, 10), sharey=True, sharex=True)

plot_competitor_points_and_fusion_lines(axs[0,0], "LowOppLowNuclear")
plot_competitor_points_and_fusion_lines(axs[0,1], "MedOppLowNuclear")
plot_competitor_points_and_fusion_lines(axs[0,2], "HighOppLowNuclear")

plot_competitor_points_and_fusion_lines(axs[1,0], "LowOpp", annotate_fusion=False)
plot_competitor_points_and_fusion_lines(axs[1,1], "MedOpp")
plot_competitor_points_and_fusion_lines(axs[1,2], "HighOpp")

plot_competitor_points_and_fusion_lines(axs[2,0], "LowOppNoNuc")
plot_competitor_points_and_fusion_lines(axs[2,1], "MedOppNoNuc")
plot_competitor_points_and_fusion_lines(axs[2,2], "HighOppNoNuc")

plot_competitor_points_and_fusion_lines(axs[3,0], "LowOppNoNucOrCCS")
plot_competitor_points_and_fusion_lines(axs[3,2], "HighOppNoNucOrCCS")

min_x = -2
max_x = 80
axs[0,0].set_xlim(min_x, max_x)
axs[0,1].set_xlim(min_x, max_x)
axs[0,2].set_xlim(min_x, max_x)
axs[0,1].set_ylim(0,800)

labelsize = 12

axsize = np.size(axs)

[ax.spines["top"].set_visible(False) for ax in axs.reshape(axsize)]
[ax.spines["right"].set_visible(False) for ax in axs.reshape(axsize)]

secax = list(np.zeros(axs.shape[0]))
for i, ax in enumerate(axs[:,2]):
    secax[i] = ax.secondary_yaxis('right', color=nice_gray)

[ax.spines["right"].set_visible(False) for ax in secax]
[ax.set_yticks([]) for ax in secax]

for ax in axs.reshape(axsize):
    ds = [10, 4]
    #plot_aries_act_point(ax)
    plot_aries_at_point(ax)
    gnx.plot_lcoe_line(ax, 50, color='gray', dashes=ds)
    gnx.plot_lcoe_line(ax, 75, color='gray', dashes=ds)

    
axbs = axs.reshape(axsize)
for i, ax in enumerate(axbs):
    ax.text(0.85, 0.9, "("+num_to_lowercase(i)+")", ha='left', va='bottom', transform=ax.transAxes)
    
axs[3,1].set_visible(False)
    
axs[2,2].secondary_yaxis('left')
[ax.secondary_xaxis('bottom') for ax in axs[1,:2]]

scen_group_labels = ["Low-cost nuclear", "Main Scenarios", "No Fission", "No Fission or CCS"]
for i, l in enumerate(scen_group_labels):
    secax[i].set_ylabel(scen_group_labels[i], color=nice_gray, labelpad=35, rotation=270)
    
scen_group_labels = ["Low market opp.", "Medium market opp.", "High market opp."]
for i, ax in enumerate(axs[0,:]):
    ax.set_title(scen_group_labels[i], color=nice_gray)

    
fig.supxlabel(myxlabel, y=0.08)
fig.supylabel(myylabel, x=0.04)

my_markers=["2", "o", "D", ">", "x"]
my_markersizes = [6, 4,  3, 3, 4]
point_reactors = [gnx.WIND, gnx.SOLAR, gnx.NUCLEAR, gnx.CCS, gnx.ZCFCC]


leg_points = [Line2D([0], [0], marker=my_markers[i], color=gnx.color_map[res], lw=0,
                          label=res, markerfacecolor=gnx.color_map[res], markersize=my_markersizes[i]) for i, res in enumerate(point_reactors)]


legend_elements = [Line2D([0], [0], color=fusion_colors[0], lw=2, label='Fusion,  10 GW'),
                   Line2D([0], [0], color=fusion_colors[1], lw=2, label='Fusion, 100 GW'),
                   Line2D([0], [0], color=fusion_colors[2], lw=2, label='Fusion, 200 GW'),
                   Line2D([0], [0], lw=0, marker='o', markersize=8, markerfacecolor='white', markeredgecolor='black', 
                          label='Example fusion plant:\nARIES-AT')]

legend_elements = legend_elements + leg_points

axs[3,2].legend(handles=legend_elements,loc=(-1.2, -0.0))

text_rot = -40
axs[1,1].text(33, 5, "$50/MWh", rotation=text_rot, color='gray', size=13)
axs[1,1].text(33+25, 5, "$75/MWh", rotation=text_rot, color='gray', size=13)

plt.savefig(gnx.product_path(chart_name+".pdf"))

plt.show()


# ### Output

# In[ ]:


cases = {"10GW": cases_10_gw, "100GW": cases_100_gw, "200GW": cases_200_gw}

scenarios_to_load = [
     "LowOppLowNuclear",
     "MedOppLowNuclear",
     "HighOppLowNuclear",
     "LowOpp",
     "MedOpp",
     "HighOpp",
     "LowOppNoNuc",
     "MedOppNoNuc",
     "HighOppNoNuc",
     "LowOppNoNucOrCCS",
     "HighOppNoNucOrCCS",
 ]


data = {scen + "_" + pen: tableaux_positions(scen, case_nums)[1]
        for scen in scenarios_to_load
        for pen, case_nums in cases.items()}

data["MarginalCostOfNetGeneration"] = tableaux_positions("MedOpp", cases_10_gw)[0]

df = pd.DataFrame(data)
a = df.to_numpy()
columns = ", ".join(df.keys())

header = """Supplementary Information Figure X:
Fusion value and competitor costs (Expanded version of main paper Figure 3).
""" + gnx.generic_data_header + \
f"""

""" + \
columns

np.savetxt(gnx.product_path(chart_name+"_lines.csv"), a, fmt="%.1f", header=header, delimiter=',')


# In[ ]:


resources_to_list = ("landbasedwind_class1_advanced",
    "offshorewind_class1_advanced_fixed",
    "landbasedwind_class1_moderate",
    "offshorewind_class1_moderate_fixed", 
    "utilitypv_class1_advanced",
    "utilitypv_class1_moderate", 
    "nuclear_nuclear_moderate",
    "naturalgas_ccs100_moderate",
    "naturalgas_ctavgcf_moderate",
    "naturalgas_ccavgcf_moderate",
)

big_list = [scen_resource_rows(scen, res) for scen in scenarios_to_load for res in resources_to_list]

no_empty_lists = [a for a in big_list if len(a) > 0]
flattened = np.array(list(chain(*no_empty_lists)))

columns = "Scenario, Resource, Variable O&M cost / ($/MWh), Annual cost / ($/MW·yr)"

header = """Figure 3:
Fusion value and competitor costs.
""" + gnx.generic_data_header + \
f"""

""" + \
columns

np.savetxt(gnx.product_path(chart_name+"_points.csv"), flattened,
           fmt="%s", delimiter=',', header=header)


# In[ ]:




