#!/usr/bin/env python
# coding: utf-8

# This script can use a large amount of memory,
# so the user might want to comment out some of the calls to make the Supplementary Data plot.

# In[1]:


import genx_fusion_analysis_base as gnx


# In[2]:


import numpy as np
import matplotlib.pyplot as plt

from matplotlib.lines import Line2D
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Patch
from matplotlib.patches import Rectangle
from matplotlib.legend_handler import HandlerLine2D, HandlerTuple
import matplotlib as mpl
import matplotlib.colors as mc

from copy import deepcopy

import colorsys
import functools
import os

import pandas as pd

import copy
import math


# In[3]:


prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']

mpl.rcParams['font.size'] = 14

η=0.4


# In[4]:


colors = gnx.colors

nice_gray = "#333333"

mpl.rcParams['font.size'] = 13
mpl.rcParams['axes.titlesize'] = 13


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = 13
mpl.rcParams['ytick.labelsize'] = 13
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 13


# In[5]:


my_scen = "MedOpp"
core_param_df = gnx.load_core_parameters_df(my_scen)
get_core_vom = functools.partial(gnx.core_vom_cost, core_param_df)


# In[6]:


def num_to_lowercase(n):
    return chr(ord("`") + 1 + n)


# In[7]:


def fusion_net_time_avg_capacity_by_zone(scen, case):
    core_params_df = gnx.load_core_parameters_df(scen)
    
    net_avg_cap_factor = gnx.get_net_avg_cap(core_params_df, case)
    
    peak_thermal_cap = gnx.ts_capacity_by_zone(scen, case)
    return peak_thermal_cap * net_avg_cap_factor


# In[8]:


def get_category_sums(df, zone_names, variety="Capacity"):
    r"""
    df: Dataframe
        capacity dataframe
    zone_names: list of str
        required to identify generator types
    """
    drop_these = [gnx.FLEX]
    
    if variety == "Capacity" or variety == "Energy":  
        col_to_sum = "EndCap"
    elif variety == "EnergyCapacity":
        col_to_sum = "EndEnergyCap"

    category_sums = dict()

    for idx, row in df.iterrows():
        cat = gnx.generator_category(row.Resource, zone_names)
        val = row[col_to_sum]
        gnx.initialize_if_not_present(category_sums, cat, 0.0)
        category_sums[cat] += val

    gnx.drop_keys(category_sums, drop_these)        
    gnx.remove_nonpositive_entries(category_sums)
    gnx.initialize_if_not_present(category_sums, "Fusion", 0.0)
    gnx.initialize_if_not_present(category_sums, gnx.CCS, 0.0)

    return category_sums


# In[9]:


@functools.cache
def capacity_sum(scen, c):
    zone_names = gnx.get_zone_names()
    df = gnx.get_capacity_df(scen, c)
    
    s = get_category_sums(df, zone_names)
    
    fus_cap = fusion_net_time_avg_capacity_by_zone(scen, c)
    s["Fusion"] = sum(fus_cap)
    
    return s

@functools.cache
def energy_capacity_sum(scen, c):
    zone_names = gnx.get_zone_names()
    df = gnx.get_capacity_df(scen, c)
    
    s = get_category_sums(df, zone_names, variety="EnergyCapacity")
    
    fusion_thermal_storage = gnx.ts_storagecap_by_zone(scen, c)
    s["Fusion"] = sum(fusion_thermal_storage)
    
    return s

@functools.cache
def capacity_sums_per_zone(scen, case):
    r"""
    """
    cap_df = gnx.get_capacity_df(scen, case)
    zone_names = gnx.get_zone_names()
    
    num_zones = len(zone_names)
    
    cat_sums = [get_category_sums(cap_df[cap_df.Zone == z], zone_names) for z in range(1, num_zones + 1)]
    
    fus_cap = fusion_net_time_avg_capacity_by_zone(scen, case)
    for z in range(num_zones):
        cat_sums[z]['Fusion'] = fus_cap[z]
        
    return cat_sums

def energy_sums(ps_df):
    keys = ps_df.keys()[1:-1]
    esums = ps_df.iloc[1].to_numpy()[1:-1]
    return pd.DataFrame(np.array([keys, esums]).T, columns=["Resource", "EndCap"])
    
def annual_energy_sum(scen, c):
    zone_names = gnx.get_zone_names()
    
    generated_energy_df = energy_sums(gnx.power_annual_sums(scen, c))
    recirculated_energy = gnx.get_total_recirc(scen, c)
    
    generated_energy = get_category_sums(generated_energy_df, zone_names)
    
    net_power = gnx.subtract_dicts(generated_energy, recirculated_energy)
    
    remove_storage_from_energy(net_power)
    return net_power

def remove_storage_from_energy(d):
    stors = ['Hydro-storage', gnx.BATTERY, gnx.METALAIR]
    gnx.drop_keys(d, stors)

@functools.cache
def annual_energy_sums_per_zone(scen, case):
    r"""
    """
    zone_names = gnx.get_zone_names()
    num_zones = len(zone_names)
    
    cap_df = gnx.get_capacity_df(scen, case)
    generated_energy_df = energy_sums(gnx.power_annual_sums(scen, case))  
    
    zonal_recirc = gnx.get_zonal_recirc(scen, case)
    zonal_recirc_as_dicts = [{'Fusion': x} for x in zonal_recirc]
      
    raw_cat_sums = [get_category_sums(generated_energy_df[cap_df.Zone == z], zone_names) for z in range(1, num_zones + 1)]
    net_energy_sums = [gnx.subtract_dicts(raw_cat_sums[i], zonal_recirc_as_dicts[i]) for i in range(num_zones)]
    
    [remove_storage_from_energy(ns) for ns in net_energy_sums]
    return net_energy_sums


# In[10]:


def dicts_to_dataframe(a):
    r"""
    a: iterable of dicts
    """
    return pd.DataFrame.from_records(a) 


# In[11]:


cap_order = [gnx.BIOMASS, gnx.HYDROELECTRIC, gnx.DISTGEN, gnx.SOLAR, gnx.WIND,  gnx.METALAIR, gnx.BATTERY, gnx.ZCF, gnx.CCS, gnx.NUCLEAR, gnx.FUSION]
energy_order = [gnx.BIOMASS, gnx.HYDROELECTRIC, gnx.DISTGEN, gnx.SOLAR, gnx.WIND,  gnx.ZCF, gnx.CCS, gnx.NUCLEAR, gnx.FUSION]
energycap_order = [gnx.HYDROELECTRIC, gnx.METALAIR, gnx.BATTERY, gnx.FUSION]

def get_sums_for_cases(scen, cases):  
    df = dicts_to_dataframe([capacity_sum(scen, c) for c in cases])
    cap_sums = df[cap_order]
    
    df = dicts_to_dataframe([annual_energy_sum(scen, c) for c in cases])
    energy_sums = df[energy_order]
    
    df = dicts_to_dataframe([energy_capacity_sum(scen, c) for c in cases])
    energycap_sums = df[energycap_order]
    return (cap_sums, energycap_sums, energy_sums,)


# In[12]:


def plot_y_difference(ax, quantities, key, **kwargs):
    end_point = kwargs.pop('end_point', -3)
    x = fusion_capacities[:end_point]
    y = quantities[key][:end_point]

    normalizer = kwargs.pop('normalizer', 1)
    return ax.plot(x, -(y[0]-y)/normalizer, **kwargs, markevery=[-1])


# In[13]:


def plot_relatives(ax, df, keykwargs, normalizer=1):
    for key, kwargs in keykwargs.items():
        pl, = plot_y_difference(ax, df, key, normalizer=normalizer, **kwargs)
        pl.set_clip_on(False)
    set_minorxticks(ax)
    
def relative_data(df, keykwargs):
    return {key: list(-df[key][0] + df[key].to_numpy()) for key in keykwargs.keys()}


# In[14]:


def add_key_prefix(dikt, prefix):
    return {prefix + key: val for key, val in dikt.items()}


# In[15]:


pes_no_storage = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
pes_med_storage = [0, 21, 22, 23, 24, 25, 26, 27, 28, 29]
mid_no_storage = [0, 10001, 10002, 10003, 10004, 10005, 10006, 10007, 10008, 10009]
mid_med_storage = [0, 10021, 10022, 10023, 10024, 10025, 10026, 10027, 10028, 10029]

opt_no_storage = [0, 101, 102, 103, 104, 105, 106, 107, 108, 109]
opt_med_storage = [0, 121, 122, 123, 124, 125, 126, 127, 128, 129]


# In[16]:


def set_minorxticks(ax):
    ax.set_xticks(fusion_capacities, minor=True)
    ax.set_xlim((0,350))

fusion_capacities = [0, 10, 20, 50, 100, 150, 200, 250, 300, 350]


# In[17]:


myxlabel = "Marginal cost of (net) generation, $/MWh"
myylabel = "Capacity cost (or value), $/MW·yr"


# In[18]:


def anncost_from_lcoe(lcoe, var):
    r"""
    LCOE: $/MWh
    var: $/MWh
    
    returns
    -------
    $/kwYr
    """
    return (lcoe - var) * 8760 / 1000


# In[19]:


relative_energy_keys = [gnx.WIND, gnx.SOLAR, gnx.NUCLEAR, gnx.CCS, gnx.ZCF]
relative_energycap_keys =  [gnx.BATTERY, gnx.METALAIR]
relative_cap_keys = relative_energy_keys


my_markers=["2", "o", "D", ">", "x", 'd', '8']
my_markersizes = [12, 7,  6, 6, 6, 7, 7]
point_reactors = [gnx.WIND, gnx.SOLAR, gnx.NUCLEAR, gnx.CCS, gnx.ZCF, gnx.BATTERY, gnx.METALAIR]

cap_legend_types = {r: {'marker': my_markers[i],
                        'ms': my_markersizes[i],
                        'color': gnx.color_map[r]} for i, r in enumerate(relative_energy_keys)}
ecap_legend_types = {r: {'color': gnx.color_map[r], 'marker': my_markers[i+5],
                         'ms': my_markersizes[i+5]} for i, r in enumerate(relative_energycap_keys)}
en_legend_types = cap_legend_types


# In[20]:


def dash_all(a_dict):
    b_dict = deepcopy(a_dict)
    for k, v in b_dict.items():
        v['ls'] = 'dashed'
    return b_dict


# In[21]:


figure_number = 5
chart_name = f"Figure_{figure_number}"


# In[22]:


scen = "MedOpp"

cap_sums_opt, energycap_sums_opt, energy_sums_opt = get_sums_for_cases(scen, opt_no_storage)
cap_sums_mid, energycap_sums_mid, energy_sums_mid = get_sums_for_cases(scen, mid_no_storage)
cap_sums_pes, energycap_sums_pes, energy_sums_pes = get_sums_for_cases(scen, pes_no_storage)

fig = plt.figure(constrained_layout=True, figsize=(9,7))
subfigs = [fig.subfigures(1, 1, wspace=0.05)]

gridspec_kw={'height_ratios':[1, 0.4, 1]}

axs1 = subfigs[0].subplots(3, 3, sharex=True, gridspec_kw=gridspec_kw, sharey='row')

   
cap_normalizer = 1e3
encap_normalizer = 1e3
ensum_normalizer = 1e6
plot_relatives(axs1[0,2], cap_sums_pes, cap_legend_types, normalizer=cap_normalizer)
plot_relatives(axs1[1,2], energycap_sums_pes, ecap_legend_types, normalizer=encap_normalizer)
plot_relatives(axs1[-1,2], energy_sums_pes, en_legend_types, normalizer=ensum_normalizer)

plot_relatives(axs1[0,1], cap_sums_mid, cap_legend_types, normalizer=cap_normalizer)
plot_relatives(axs1[1,1], energycap_sums_mid, ecap_legend_types, normalizer=encap_normalizer)
plot_relatives(axs1[-1,1], energy_sums_mid, en_legend_types, normalizer=ensum_normalizer)

plot_relatives(axs1[0,0], cap_sums_opt, cap_legend_types, normalizer=cap_normalizer)
plot_relatives(axs1[1,0], energycap_sums_opt, ecap_legend_types, normalizer=encap_normalizer)
plot_relatives(axs1[-1,0], energy_sums_opt, en_legend_types, normalizer=ensum_normalizer)


for ax in axs1.reshape(9):
    ax.hlines([0], 0, 350, lw=0.5, color='gray')
    ax.set_xlim(0, 200)
    [ax.spines[side].set_visible(False) for side in ["top", "right"]]

axs1[0,0].set_ylim(-250, 50)
axs1[1,0].set_ylim(-800, 250)
axs1[-1,0].set_ylim(-700, 150)

[ax.set_ylabel(yl) for ax, yl in zip(axs1[:,0], ("Δ capacity / GW", "Δ storage / GWh", "Δ production / TWh"))]
[ax.set_title(t) for ax, t in zip(axs1[0,:], gnx.fusion_reactor_names)]

fusion_markers = ["P", "*", "s"]
[ax.plot([150], [67], marker=fusion_markers[i], ms=8, color=colors[i], clip_on=False) for i, ax in enumerate(axs1[0, :])]

#subfigs[0].supxlabel("Marginal cost of (net) generation / ($/MWh)", size=14)
subfigs[0].supxlabel("Fusion capacity / GW", size=14)

#axs0.text(0.98, 0.98, '(a)', ha='right', va='top', transform=axs0.transAxes)

for i, ax in enumerate(axs1.reshape(9)):
    ax.text(0.04, 0.98, "("+num_to_lowercase(i)+")", ha='left', va='top', transform=ax.transAxes)


my_markers = ["D", "o", "2", ">", "x", 'd', '8']
my_markersizes = [6, 8, 15,   8, 10, 7, 7]
point_reactors = [ gnx.NUCLEAR, gnx.SOLAR, gnx.WIND, gnx.CCS, gnx.ZCF, gnx.BATTERY, gnx.METALAIR]
leg_points = [Line2D([0], [0], marker=my_markers[i], color=gnx.color_map[res], lw=0,
                          label=res, markerfacecolor=gnx.color_map[res], markersize=my_markersizes[i]) for i, res in enumerate(point_reactors)]

legend_elements = leg_points

subfigs[0].legend(handles=legend_elements, bbox_to_anchor=(0, 1, 1, 0), loc="lower center", ncol=4)

plt.savefig(gnx.product_path(f"{chart_name}.pdf"), bbox_inches='tight')

plt.show()


# ### Output of data

# In[23]:


def format_data_for_scenario(scen):
    
    cap_sums_opt, energycap_sums_opt, energy_sums_opt = get_sums_for_cases(scen, opt_no_storage)
    cap_sums_mid, energycap_sums_mid, energy_sums_mid = get_sums_for_cases(scen, mid_no_storage)
    cap_sums_pes, energycap_sums_pes, energy_sums_pes = get_sums_for_cases(scen, pes_no_storage)  


    part_c = add_key_prefix(relative_data(cap_sums_pes, cap_legend_types), "(c) PessimisticFusion Capacity ")
    part_b = add_key_prefix(relative_data(cap_sums_mid, cap_legend_types), "(b) MidrangeFusion Capacity ")
    part_a = add_key_prefix(relative_data(cap_sums_opt, cap_legend_types), "(a) OptimisticFusion Capacity ")

    part_f = add_key_prefix(relative_data(energycap_sums_pes, ecap_legend_types), "(f) PessimisticFusion Storage ")
    part_e = add_key_prefix(relative_data(energycap_sums_mid, ecap_legend_types), "(e) MidrangeFusion Storage ")
    part_d = add_key_prefix(relative_data(energycap_sums_opt, ecap_legend_types), "(d) OptimisticFusion Storage ")

    part_i = add_key_prefix(relative_data(energy_sums_pes, en_legend_types), "(i) PessimisticFusion Energy Production ")
    part_h = add_key_prefix(relative_data(energy_sums_mid, en_legend_types), "(h) MidrangeFusion Energy Production ")
    part_g = add_key_prefix(relative_data(energy_sums_opt, en_legend_types), "(h) OptimisticFusion Energy Production ")

    x = {"Fusion capacities": fusion_capacities}
    all_data = x | part_a | part_b | part_c | part_d | part_e | part_f | part_g | part_h | part_i

    
    is_medopp = scen == "MedOpp"
    withstorage = scen=="LowOpp" or scen == "HighOpp" or is_medopp or scen == "CRMedOpp"
    if withstorage:
        stor_mark = "WithMidCostStorage"
    
        cap_sums_mid_med, energycap_sums_mid_med, energy_sums_mid_med = get_sums_for_cases(scen, mid_med_storage)
        
        part_bs = add_key_prefix(relative_data(cap_sums_mid, cap_legend_types),
                                 f"(b) {stor_mark} MidrangeFusion Capacity ")
        part_es = add_key_prefix(relative_data(energycap_sums_mid, ecap_legend_types),
                                 f"(e) {stor_mark} MidrangeFusion Storage ")
        part_hs = add_key_prefix(relative_data(energy_sums_mid, en_legend_types),
                                 f"(h) {stor_mark} MidrangeFusion Energy Production ")
        all_data = all_data | part_bs | part_es | part_hs

    if is_medopp:
        cap_sums_pes_med, energycap_sums_pes_med, energy_sums_pes_med = get_sums_for_cases(scen, pes_med_storage)
        cap_sums_opt_med, energycap_sums_opt_med, energy_sums_opt_med = get_sums_for_cases(scen, opt_med_storage)
        part_cs = add_key_prefix(relative_data(cap_sums_pes, cap_legend_types), f"(c) {stor_mark} PessimisticFusion Capacity ")
        part_as = add_key_prefix(relative_data(cap_sums_opt, cap_legend_types), f"(a) {stor_mark} OptimisticFusion Capacity ")
        part_fs = add_key_prefix(relative_data(energycap_sums_pes, ecap_legend_types), f"(f) {stor_mark} PessimisticFusion Storage ")
        part_ds = add_key_prefix(relative_data(energycap_sums_opt, ecap_legend_types), f"(d) {stor_mark} OptimisticFusion Storage ")
        part_is = add_key_prefix(relative_data(energy_sums_pes, en_legend_types), f"(i) {stor_mark} PessimisticFusion Energy Production ")
        part_gs = add_key_prefix(relative_data(energy_sums_opt, en_legend_types), f"(h) {stor_mark} OptimisticFusion Energy Production ")
        all_data = all_data | part_as | part_cs | part_ds | part_fs | part_gs | part_is    

    
    columns = ", ".join(list(all_data.keys()))

    a = pd.DataFrame(all_data).to_numpy()
    
    return a, columns
    
def save_data_for_scenario(scen):
    a, columns = format_data_for_scenario(scen)

    header = f"""{chart_name}:
    
Substutited generation capacity, energy storage capacity, and energy produced 
as the fusion capacity increases.
""" + gnx.generic_data_header + \
f"""
This is for the {scen} scenario.
Optimistic, Midrange, and Pessimistic refer to the three reference fusion reactors.

Fusion capacities are in GW.
Capacities in MW. 
Storage capacities are in MWh.
Energy Production in MWh.

Negative values mean that less capacity/production for resource type X exists 
in the case with the given fusion capacity than in the case with 0 fusion capacity.
Letters (a)-(i) refer to the original subfigure label for the data in this column.

""" + \
    columns
    
    np.savetxt(gnx.product_path(f"{chart_name}.csv"), a, header=header, delimiter=',', fmt="%.1f")
    
    
def save_data_for_supplemental_scenario(scen):
    a, columns = format_data_for_scenario(scen)

    header = f"""Supplemental data: displacement of other resources in {scen} scenario.
    
Substutited generation capacity, energy storage capacity, and energy produced 
as the fusion capacity increases.
""" + gnx.generic_data_header + \
f"""
Optimistic, Midrange, and Pessimistic refer to the three reference fusion reactors.

Fusion capacities are in GW.
Capacities in MW. 
Storage capacities are in MWh.
Energy Production in MWh.

Negative values mean that less capacity/production for resource type X exists 
in the case with the given fusion capacity than in the case with 0 fusion capacity.
Letters (a)-(i) refer to the original subfigure label for the data in this column.

Additional columns (if any) with the notation "WithMidCostStorage" are for cases with $22/kWh molten salt storage.
These cases were essentially only run for the main scenarios.

""" + \
    columns
    
    np.savetxt(gnx.product_path("SUPP_displacement_"+scen+".csv"), a, header=header, delimiter=',', fmt="%.1f")


# In[ ]:


save_data_for_scenario("MedOpp")


# # Suppementary Data

# In[24]:


def fusion_substitutes_plot(scen, ymaxes=(235,800, 700), my_xmax=350):
    
    my_scen = scen

    cap_sums_opt, energycap_sums_opt, energy_sums_opt = get_sums_for_cases(scen, opt_no_storage)
    cap_sums_mid, energycap_sums_mid, energy_sums_mid = get_sums_for_cases(scen, mid_no_storage)
    cap_sums_pes, energycap_sums_pes, energy_sums_pes = get_sums_for_cases(scen, pes_no_storage)
    
    is_medopp = scen == "MedOpp"
    withstorage = scen=="LowOpp" or scen == "HighOpp" or is_medopp or scen == "CRMedOpp"
    if withstorage:
        cap_sums_mid_med, energycap_sums_mid_med, energy_sums_mid_med = get_sums_for_cases(scen, mid_med_storage)

    if is_medopp:
        cap_sums_pes_med, energycap_sums_pes_med, energy_sums_pes_med = get_sums_for_cases(scen, pes_med_storage)
        cap_sums_opt_med, energycap_sums_opt_med, energy_sums_opt_med = get_sums_for_cases(scen, opt_med_storage)

        

    fig = plt.figure(constrained_layout=True, figsize=(9,10))
    subfigs = [fig.subfigures(1, 1, wspace=0.05)]

    gridspec_kw={'height_ratios':[1, 0.5, 1]}

    axs1 = subfigs[0].subplots(3, 3, sharex=True, gridspec_kw=gridspec_kw, sharey='row')

    relative_energy_keys = [gnx.WIND, gnx.SOLAR, gnx.NUCLEAR, gnx.CCS, gnx.ZCF]
    relative_energycap_keys =  [gnx.BATTERY, gnx.METALAIR]
    relative_cap_keys = relative_energy_keys

    def dash_all(a_dict):
        b_dict = deepcopy(a_dict)
        for k, v in b_dict.items():
            v['ls'] = 'dashed'
        return b_dict

    my_markers=["2", "o", "D", ">", "x", 'd', '8']
    my_markersizes = [12, 7,  6, 6, 6, 7, 7]
    point_reactors = [gnx.WIND, gnx.SOLAR, gnx.NUCLEAR, gnx.CCS, gnx.ZCF, gnx.BATTERY, gnx.METALAIR]

    cap_legend_types = {r: {'end_point': None, 'marker': my_markers[i], 'ms': my_markersizes[i], 'color': gnx.color_map[r]} for i, r in enumerate(relative_energy_keys)}
    ecap_legend_types = {r: {'end_point': None, 'color': gnx.color_map[r], 'marker': my_markers[i+5],
                             'ms': my_markersizes[i+5]} for i, r in enumerate(relative_energycap_keys)}
    en_legend_types = cap_legend_types

    cap_normalizer = 1e3
    encap_normalizer = 1e3
    ensum_normalizer = 1e6
    plot_relatives(axs1[0,2], cap_sums_pes, cap_legend_types, normalizer=cap_normalizer)
    plot_relatives(axs1[1,2], energycap_sums_pes, ecap_legend_types, normalizer=encap_normalizer)
    plot_relatives(axs1[-1,2], energy_sums_pes, en_legend_types, normalizer=ensum_normalizer)

    plot_relatives(axs1[0,1], cap_sums_mid, cap_legend_types, normalizer=cap_normalizer)
    plot_relatives(axs1[1,1], energycap_sums_mid, ecap_legend_types, normalizer=encap_normalizer)
    plot_relatives(axs1[-1,1], energy_sums_mid, en_legend_types, normalizer=ensum_normalizer)

    plot_relatives(axs1[0,0], cap_sums_opt, cap_legend_types, normalizer=cap_normalizer)
    plot_relatives(axs1[1,0], energycap_sums_opt, ecap_legend_types, normalizer=encap_normalizer)
    plot_relatives(axs1[-1,0], energy_sums_opt, en_legend_types, normalizer=ensum_normalizer)

    if withstorage:
        cap_stor_legend_types = dash_all(cap_legend_types)
        ecap_stor_legend_types = dash_all(ecap_legend_types)
        en_stor_legend_types = dash_all(en_legend_types)
    
        plot_relatives(axs1[0,1], cap_sums_mid_med, cap_stor_legend_types, normalizer=cap_normalizer)
        plot_relatives(axs1[1,1], energycap_sums_mid_med, ecap_stor_legend_types, normalizer=encap_normalizer)
        plot_relatives(axs1[-1,1], energy_sums_mid_med, en_stor_legend_types, normalizer=ensum_normalizer)

    if is_medopp:
        plot_relatives(axs1[0,2], cap_sums_pes_med, cap_stor_legend_types, normalizer=cap_normalizer)
        plot_relatives(axs1[1,2], energycap_sums_pes_med, ecap_stor_legend_types, normalizer=encap_normalizer)
        plot_relatives(axs1[-1,2], energy_sums_pes_med, en_stor_legend_types, normalizer=ensum_normalizer)
        plot_relatives(axs1[0,0], cap_sums_opt_med, cap_stor_legend_types, normalizer=cap_normalizer)
        plot_relatives(axs1[1,0], energycap_sums_opt_med, ecap_stor_legend_types, normalizer=encap_normalizer)
        plot_relatives(axs1[-1,0], energy_sums_opt_med, en_stor_legend_types, normalizer=ensum_normalizer)

        
    for ax in axs1.reshape(9):
        ax.hlines([0], 0, 350, lw=0.5, color='gray')
        ax.set_xlim(0, my_xmax)
        [ax.spines[side].set_visible(False) for side in ["top", "right"]]

    axs1[0,0].set_ylim(-ymaxes[0], 100)
    axs1[1,0].set_ylim(-ymaxes[1], 500)
    axs1[-1,0].set_ylim(-ymaxes[2], 150)

    [ax.set_ylabel(yl) for ax, yl in zip(axs1[:,0], ("Δ capacity / GW", "Δ storage / GWh", "Δ production / TWh"))]
    [ax.set_title(t) for ax, t in zip(axs1[0,:], gnx.fusion_reactor_names)]

   
    subfigs[0].supxlabel("Fusion capacity / GW", size=14)

    for i, ax in enumerate(axs1.reshape(9)):
        ax.text(0.04, 1.0, "("+num_to_lowercase(i)+")", ha='left', va='top', transform=ax.transAxes)


    my_markers = ["D", "o", "2", ">", "x", 'd', '8']
    my_markersizes = [6, 8, 15,   8, 10, 7, 7]
    point_reactors = [ gnx.NUCLEAR, gnx.SOLAR, gnx.WIND, gnx.CCS, gnx.ZCF, gnx.BATTERY, gnx.METALAIR]
    leg_points = [Line2D([0], [0], marker=my_markers[i], color=gnx.color_map[res], lw=0,
                              label=res, markerfacecolor=gnx.color_map[res],
                         markersize=my_markersizes[i]) for i, res in enumerate(point_reactors)]

    legend_elements = leg_points

    subfigs[0].legend(handles=legend_elements, bbox_to_anchor=(0, 1, 1, 0), loc="lower center", ncol=4)

    plt.savefig(gnx.product_path("SUPP_displacement_" + scen + ".pdf"), bbox_inches='tight')

    save_data_for_supplemental_scenario(scen)
    
    plt.show()
    
ymaxes=(800, 2500, 1600)


# In[ ]:


fusion_substitutes_plot("LowOpp", ymaxes=ymaxes)

fusion_substitutes_plot("MedOpp", ymaxes=ymaxes)

fusion_substitutes_plot("HighOpp", ymaxes=ymaxes)


# In[ ]:


fusion_substitutes_plot("CRLowOpp", ymaxes=ymaxes)

fusion_substitutes_plot("CRMedOpp", ymaxes=ymaxes)

fusion_substitutes_plot("CRHighOpp", ymaxes=ymaxes)


# In[25]:


fusion_substitutes_plot("HighOppNoNucOrCCS", ymaxes=ymaxes)

fusion_substitutes_plot("LowOppLowNuclear", ymaxes=ymaxes)

fusion_substitutes_plot("MedOppLowNuclear", ymaxes=ymaxes)

fusion_substitutes_plot("HighOppLowNuclear", ymaxes=ymaxes)


# In[ ]:




