#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx


# In[ ]:


import numpy as np
import matplotlib.pyplot as plt

import matplotlib as mpl


# In[ ]:


nice_gray = "#333333"


print_text_size = 8

mpl.rcParams['font.size'] = print_text_size
mpl.rcParams['axes.titlesize'] = print_text_size

mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = print_text_size
mpl.rcParams['figure.titlesize'] = print_text_size


# In[ ]:


figsize = (6,6)


# In[ ]:


def get_values_and_normalizers(scen, base_case, f):
    cases_without_storage = np.arange(0,6)
    storage_cases = np.arange(6,48)
    normalizers = f(scen, base_case + cases_without_storage)
    values = f(scen, base_case + storage_cases)
    return normalizers, values


# # Storage plots: Series 20000

# In[ ]:


s20k =20*1000

scen_low = "LowOpp"
scen_med = "MedOpp"
scen_high = "HighOpp"


# ## Data fill-in

# In[ ]:


def fancy_ticks_xaxis(ax):
    ax.set_xlim(4, 30)
    ax.set_xticks([4,12,26], minor=False)
    ax.set_xticks([10,20], minor=True)
    
def row_opp_label(ax, r, s, xloc=None):
    if xloc is None:
        xloc = ax[r,2].get_xlim()[1] + 1
    ax[r,2].text(xloc, 22, s, size=print_text_size, rotation=270, verticalalignment='center')


# In[ ]:


mid_vom = 3
mid_pulse_length = 4
mid_dwell = 0.15
mid_pass_recirc = 0.1
mid_act_recirc = 0.1
mid_start_energy = 0.025
mid_η = 0.4
mid_gen_vom = 1.74


# In[ ]:


mid_vomcostpermwhth = [0.5, 2,3,5,8,13]
mid_totvom = [gnx.totalVOMcost(x, mid_η, mid_pulse_length, mid_dwell, mid_pass_recirc, mid_act_recirc, mid_start_energy) for x in mid_vomcostpermwhth]


# In[ ]:


data_shape = (6,7)

series20k_storagecosts_base = list(gnx.ann_to_capex(0.858 * 2**np.linspace(0, 2, 7)))
series20k_vomcosts_base = mid_totvom
series20k_storagecosts, series20k_vomcosts = np.meshgrid(np.r_[series20k_storagecosts_base], 
                                                         series20k_vomcosts_base)


# In[ ]:


coordinates_header = f"""

Each subplot is made from a 6 x 7 grid of data with coordinates.

X coordinates: {np.round(series20k_vomcosts_base, 3)} $/MWh plant VOM cost
Y coordinates: {np.round(series20k_storagecosts_base,3)} $/kWh storage capacity capital cost

Note that the order of the y coordinates is 'flipped' here vs how they visually appear in the corresponding plots.
(The first rows are the lowest y coordinate.)

(a) (b) (c) corresponds to the first row of subplots, (d) (e) (f) is the second row, etc.

(a):  20 GW fusion capacity, LowOpp
(b): 100 GW fusion capacity, LowOpp
(c): 200 GW fusion capacity, LowOpp
(d):  20 GW fusion capacity, MedOpp
(e): 100 GW fusion capacity, MedOpp
(f): 200 GW fusion capacity, MedOpp
(g):  20 GW fusion capacity, HighOpp
(h): 100 GW fusion capacity, HighOpp
(i): 200 GW fusion capacity, HighOpp"""


# In[ ]:


def plot_scenario_capacity_row(f, ax, row, scen, base_case):
    f(ax[row, 0], scen, base_case)
    f(ax[row, 1], scen, base_case + 100)
    f(ax[row, 2], scen, base_case + 200)

def plot_scenario_reactor_row_2(f, ax, row, scen, base_case):
    f(ax[row, 0], scen, base_case + 2000 + 100)
    f(ax[row, 1], scen, base_case + 100)
    f(ax[row, 2], scen, base_case + 1000 + 100)


# In[ ]:


def set_cap_col_titles(ax):
    ax[0,0].set_title("Net fusion capacity 20 GW")
    ax[0,1].set_title("50 GW")
    ax[0,2].set_title("100 GW")
    
def set_design_col_titles(ax):
    ax[0,0].set_title("Pessimistic reactor")
    ax[0,1].set_title("Mid-range reactor")
    ax[0,2].set_title("Optimistic reactor")


# In[ ]:


def set_block_ticks(ax):
    [fancy_ticks_xaxis(a[0]) for a in ax.reshape(1,ax.size)]
    
    my_ylabels = [11, 22 ,  44]
    ax[0,0].set_yticks([], minor=True)
    ax[0,0].set_yticks(my_ylabels, labels=my_ylabels, minor=False)


# In[ ]:


def nine_blocks_cap_scen(f, ax, base):
    plot_scenario_capacity_row(f, ax, 0, scen_low, base)
    plot_scenario_capacity_row(f, ax, 1, scen_med, base)
    plot_scenario_capacity_row(f, ax, 2, scen_high, base)
    
    set_cap_col_titles(ax)
    set_block_ticks(ax)
        
    row_opp_label(ax, 0, "Low fusion market opp.")
    row_opp_label(ax, 1, "Medium opp.")
    row_opp_label(ax, 2, "High opp.")


# In[ ]:


def nine_blocks_reactor_scen_2(f, ax, base):
    #plot_scenario_reactor_row_2(f, ax, 0, scen_low, base)
    plot_scenario_reactor_row_2(f, ax, 1, scen_med, base)
    plot_scenario_reactor_row_2(f, ax, 2, scen_high, base)
    
    set_design_col_titles(ax)
    set_block_ticks(ax)
        
    row_opp_label(ax, 0, "Low opp.")
    row_opp_label(ax, 1, "Medium opp.")
    row_opp_label(ax, 2, "High opp.")


# In[ ]:


def my_contour(ax, v, my_levels, vmin, vmax):
    x = series20k_vomcosts
    y = series20k_storagecosts
    CS = ax.contourf(x, y, v, my_levels, vmin=vmin, vmax=vmax)
    CS2 = ax.contour(x, y, v, my_levels, vmin=vmin, vmax=vmax, colors=None)

    clabel = ax.clabel(CS2, my_levels[1:-1], inline=True, fontsize=print_text_size, colors='black')
    [txt.set_bbox(dict(facecolor='white', boxstyle="round", alpha=0.2, edgecolor='none', pad=0.2)) for txt in clabel]
    #[txt.set_backgroundcolor('white') for txt in clabel]
    CS3 = ax.contour(x, y, v, my_levels, vmin=vmin, vmax=vmax, colors=None)

    ax.set_yscale('log')


# In[ ]:


def array_as_csv_string(a, digits=4):
    out = []
    for row in a:
        els = [str(round(el, digits)) for el in row]
        out.append(",".join(els))
    return "\n".join(out)

def num_to_lowercase(n):
    return chr(ord("`") + 1 + n)

def data_blocks_to_string(d_blocks, **kwargs):
    out = []
    for i, a in enumerate(d_blocks):
        out.append("# (" + num_to_lowercase(i) + ")")
        out.append(array_as_csv_string(a.T, **kwargs))
    return "\n".join(out)

def prepend_hash(string):
    lines = string.split('\n')
    commented_lines = ['# ' + l for l in lines]
    return "\n".join(commented_lines)


# In[ ]:


marginal_cost_of_generation = "Marginal cost of net generation, $/MWh"


# In[ ]:


def data_blocks():
    scens = ["LowOpp", "MedOpp", "HighOpp"]
    base_cases = [s20k, s20k+100, s20k+200]
    return [block_data(scen, bc) for scen in scens for bc in base_cases ]


# In[ ]:


fig, ax = plt.subplots(3,3, figsize=figsize, sharex=True, sharey=True)
my_levels = np.arange(0.975,1.175,0.025)
my_levels = np.append(my_levels, 1.001)
my_levels.sort()

contour_min = min(my_levels)
contour_max = max(my_levels)

def reshape_and_normalize(values, normalizers):
    v = np.reshape(values, data_shape)
    return np.maximum((v.T / normalizers).T, 1.00001)
    
def plot_with_labels(ax, v):
    my_contour(ax, v, my_levels, contour_min, contour_max)
    
def block_data(scen, base_case):
    norms, vals = get_values_and_normalizers(scen, base_case, gnx.grab_duals)
    v = reshape_and_normalize(vals, norms)
    return v
    
def plot_block(ax, scen, base_case):
    v = block_data(scen, base_case)
    plot_with_labels(ax, v)

nine_blocks_cap_scen(plot_block, ax, s20k)
#nine_blocks_reactor_scen_2(plot_block, ax, s20k)


#fig.suptitle("Fractional enhancement of core's value with storage")
fig.supxlabel(marginal_cost_of_generation)
fig.supylabel("Storage capacity cost $/kWh")

file_name = "SUPP_storage_fractional-value-enhancement_Midrange"
plt.tight_layout()
plt.savefig(gnx.product_path(file_name + '.pdf'))
header = f"""Supplemental: Fractional value enhancement for mid-range cores with storage.

Values given are unitless: (Value of core with storage) / (Value of core with storage).
See the methods section and supplementary information for a precise definition of the plant's "core".
""" + \
gnx.generic_data_header + coordinates_header
out = prepend_hash(header) + '\n' + data_blocks_to_string(data_blocks())

with open(gnx.product_path(file_name + ".csv"), 'w') as f:
    f.write(out)


# In[ ]:


def reshape_and_subtract(values, subtractors):
    v = np.reshape(values, data_shape)
    return (v.T - subtractors).T

fig, ax = plt.subplots(3,3, figsize=(6,6), sharex=True, sharey=True)
my_levels = np.arange(0, 1101, 100)
my_levels = np.append(my_levels, [5, 50])
my_levels.sort()

contour_min = min(my_levels)
contour_max = max(my_levels) - 100

def plot_with_labels(ax, v):
    v = np.maximum(v, 0)
    my_contour(ax, v, my_levels, contour_min, contour_max)
    
def block_data(scen, base_case):
    norms, vals = get_values_and_normalizers(scen, base_case, gnx.core_value_capex)
    v = reshape_and_subtract(vals, norms)
    return v
    
def plot_block(ax, scen, base_case):
    v = block_data(scen, base_case)
    plot_with_labels(ax, v)
    
nine_blocks_cap_scen(plot_block, ax, s20k)

#fig.suptitle('Increase in core value with storage $/kw, Mid-range reactor')
fig.supylabel("Storage capacity cost $/kWh")
fig.supxlabel(marginal_cost_of_generation)

file_name = "Figure_6"
plt.tight_layout()
plt.savefig(gnx.product_path(file_name + '.pdf'))
plt.show()
header = f"""Figure 6: Core value enhancement for mid-range plants with storage.

Numbers are: (Value of core with storage) - (Value of core without storage)
and are terms of the capitalized cost in 2019$/MW.
See the methods section and supplementary information for a precise definition of the plant's "core".
""" + \
gnx.generic_data_header + coordinates_header
out = prepend_hash(header) + '\n' + data_blocks_to_string(data_blocks(), digits=1)

with open(gnx.product_path(file_name + ".csv"), 'w') as f:
    f.write(out)


# In[ ]:


def my_reshape(values, subtractors):
    v = np.reshape(values, data_shape)
    return v

fig, ax = plt.subplots(3,3, figsize=figsize, sharex=True, sharey=True)
my_levels = np.arange(0, 8001, 500)
contour_min = min(my_levels)
contour_max = max(my_levels) - 100

def plot_with_labels(ax, v):
    v = np.maximum(v, 0)
    my_contour(ax, v, my_levels, contour_min, contour_max)
    
def block_data(scen, base_case):
    norms, vals = get_values_and_normalizers(scen, base_case, gnx.core_value_capex)
    v = my_reshape(vals, norms)
    return v
    
def plot_block(ax, scen, base_case):
    v = block_data(scen, base_case)
    plot_with_labels(ax, v)
            
nine_blocks_cap_scen(plot_block, ax, s20k)

#fig.suptitle('Mid-range reactor core value with storage, $/kW')
fig.supylabel("Storage capacity cost $/kWh", x=0.06)
fig.supxlabel(marginal_cost_of_generation, y=0.05)

file_name = "SUPP_storage_core_value"
plt.savefig(gnx.product_path(file_name + '.pdf'))
plt.show()
header = f"""Supplemental: Core value of mid-range plants with storage.

Values are the capitalized cost of the core, in 2019$/MW.
See the methods section and supplementary information for a precise definition of the plant's "core".
""" + \
gnx.generic_data_header + coordinates_header
out = prepend_hash(header) + '\n' + data_blocks_to_string(data_blocks(), digits=1)

with open(gnx.product_path(file_name + ".csv"), 'w') as f:
    f.write(out)

plt.show()


# In[ ]:


fig, ax = plt.subplots(3,3, figsize=figsize, sharex=True, sharey=True)
my_levels = [0.9,1.0,1.1,1.2,1.3,1.4,1.5, 1.6]
contour_min = 0.9
contour_max = 1.55

def plot_with_labels(ax, values):
    v = np.reshape(values, data_shape)
    my_contour(ax, v, my_levels, contour_min, contour_max)
       
def block_data(scen, base_case):
    _, vals = get_values_and_normalizers(scen, base_case, gnx.grab_generation_ratios)
    v = np.reshape(vals, data_shape)
    return v
    
def plot_block(ax, scen, base_case):
    v = block_data(scen, base_case)
    plot_with_labels(ax, v)

nine_blocks_cap_scen(plot_block, ax, s20k)

#fig.suptitle('Ratio of generator capacity to core capacity')
fig.supylabel("Storage capacity cost $/kWh/yr")
fig.supxlabel(marginal_cost_of_generation)

file_name = 'SUPP_storage_generator-ratios_MidRange'
plt.tight_layout()
plt.savefig(gnx.product_path(file_name + '.pdf'))
plt.show()
header = f"""Supplemental: Ratio of generator capacity to long-run plant capacity

Values are the ratio of the peak generator capacity to that of the long-run capacity of the plant.
""" + \
gnx.generic_data_header + coordinates_header
out = prepend_hash(header) + '\n' + data_blocks_to_string(data_blocks(), digits=3)

with open(gnx.product_path(file_name + ".csv"), 'w') as f:
    f.write(out)

plt.show()


# In[ ]:


fig, ax = plt.subplots(3,3, figsize=figsize, sharex=True, sharey=True)
my_levels = [0, 0.15, 1.5,3,4.5,6,7.5,9,10.5,12,13.5,15, 18]
contour_min = 0
contour_max = 15

def plot_with_labels(ax, v):
    my_contour(ax, v, my_levels, contour_min, contour_max)
    
def block_data(scen, base_case):
    _, vals = get_values_and_normalizers(scen, base_case, gnx.grab_duration)
    v = np.reshape(vals, data_shape)
    return v
    
def plot_block(ax, scen, base_case):
    v = block_data(scen, base_case)
    plot_with_labels(ax, v)

nine_blocks_cap_scen(plot_block, ax, s20k)

fig.supylabel("Storage capacity cost $/kWh")
fig.supxlabel(marginal_cost_of_generation)

file_name = "SUPP_storage_duration_Midrange"
plt.tight_layout()
plt.savefig(gnx.product_path(file_name + '.pdf'))
plt.show()
header = f"""Supplemental: Storage duration in hours for Mid-range plants

Values are the ratio of storage energy capacity to the peak core thermal capacity
, in hours.
""" + \
gnx.generic_data_header + coordinates_header
out = prepend_hash(header) + '\n' + data_blocks_to_string(data_blocks(), digits=3)

with open(gnx.product_path(file_name + ".csv"), 'w') as f:
    f.write(out)

