#!/usr/bin/env python
# coding: utf-8

# In[5]:


import genx_fusion_analysis_base as gnx


# In[6]:


import numpy as np
import matplotlib.pyplot as plt

from matplotlib.lines import Line2D
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Patch
from matplotlib.patches import Rectangle

import matplotlib.patches as mpatches

import matplotlib as mpl

from copy import deepcopy

import colorsys
import functools
import os

import pandas as pd
from scipy.interpolate import interp1d


# Note: these cases have 6 hour ... 1 hour up/down times.

# In[7]:


colors = gnx.colors

nice_gray = "#333333"

#mpl.rcParams['font.family'] = 'sans-serif'
mpl.rcParams['font.size'] = 8
mpl.rcParams['axes.titlesize'] = 10


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = 8
mpl.rcParams['ytick.labelsize'] = 8
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 8


# In[8]:


def generator_per_netmwe_10k(scen, indices):
    gen = []
    for i in indices:
        if i >= 10000 and i < 20000 or i==4 or i==104:
            divisor = 100e3
        elif i>=20000 and i < 20100:
            divisor = 20e3
        elif i>=20100 and i < 20200:
            divisor = 50e3
        elif i>=20200 and i < 20300:
            divisor = 100e3
        elif i>=20300 and i < 20400:
            divisor = 150e3
        elif i>=20400 and i < 20500:
            divisor = 200e3
        elif i>=20900 and i < 20906:
            divisor = 10e3
        else:
            raise Exception("Error, not programmed yet")
        gen.append(gnx.grab_data('generation', scen, [i])[0] / divisor)
    return np.array(gen)

def additional_generation_cost(scen, indices):
    return gnx.base_gen_cost * generator_per_netmwe_10k(scen, indices)


# In[9]:


η = 0.4

mid_vom = 3
mid_pulse_length = 4
mid_dwell = 0.15
mid_pass_recirc = 0.1
mid_act_recirc = 0.1
mid_start_energy = 0.025

def tableaux_positions(scen, cases):
    cases_vomcostpermwhth = [0.5, 2, 3, 5, 8, 13]
    cases_duals = gnx.grab_duals(scen, cases)
    cases_vom = [gnx.totalVOMcost(x, gnx.base_η, mid_pulse_length, mid_dwell, mid_pass_recirc, mid_act_recirc, mid_start_energy) for x in cases_vomcostpermwhth]
    cases_totcost = cases_duals + additional_generation_cost(scen, cases)
    return cases_vom, cases_totcost/1e3
    


# In[10]:


r = range(9, 45, 7)


# In[11]:


nostor_increments = range(6)
midstor_increments = range(9, 45, 7)
cases_10_gw = [20900 + i for i in nostor_increments]

cases_50_gw = [20100 + i for i in nostor_increments]
cases_100_gw = [20200 + i for i in nostor_increments]

cases_50_gw_stor = [20100 + i for i in midstor_increments]
cases_100_gw_stor = [20200 + i for i in midstor_increments]

cases_200_gw = [20400 + i for i in nostor_increments]


# In[12]:


cases_200_gw_mustrun = [20400 + i for i in nostor_increments]


# In[13]:


def num_to_lowercase(n):
    return chr(ord("`") + 1 + n)


# # Calculation of effective LCOE

# In[14]:


def fusion_resource_name(z):
    return (gnx.ZONE_NAMES[z]+'_fusion').lower()

def fusion_generation_in_zone(power_df, zone):
    resource_name = fusion_resource_name(zone)
    generated = power_df[resource_name]
    return generated.to_numpy()

def fusion_recirculation_in_zone(recirc_df, zone):
    resource_name = fusion_resource_name(zone)
    recirc = recirc_df[resource_name]
    return recirc.to_numpy()

def net_fusion_generation_in_zone(power_df, recirc_df, zone):
    gen = fusion_generation_in_zone(power_df, zone)
    rec = fusion_recirculation_in_zone(recirc_df, zone)
    return gen - rec

def fusion_core_power_in_zone(corepower_df, zone):
    resource_name = fusion_resource_name(zone)
    generated = corepower_df[resource_name]
    return generated.to_numpy()

def fusion_generation_capacity_in_zone(df, zone):
    resource_name = fusion_resource_name(zone)
    cap = df[df.Resource == resource_name].EndCap
    return cap.to_numpy()[0]


# In[15]:


def data_for_reactor_in_zone(scen, case, z):
    
    param_df =  gnx.load_core_parameters_df(scen)
    capacity_df = gnx.get_capacity_df(scen, case)
    corepower_df = gnx.fusion_corepower_hourly(scen, case)
    power_df = gnx.power_hourly(scen, case)
    recirc_df = gnx.fusion_recirc(scen, case)
    active_fraction = gnx.get_active_fraction(param_df, case)

    ts_capacity = gnx.ts_capacity_by_zone(scen, case)[z]
    gross_gen_cap = fusion_generation_capacity_in_zone(capacity_df, z)
    expected_gen_cap = gnx.get_long_run_gen_to_core_ratio(param_df, case)
    net_avg_cap = gnx.get_net_avg_cap(param_df, case)
    long_run_average_power = ts_capacity * net_avg_cap
    min_net_power_fraction = gnx.get_minimum_net_power_fraction(param_df, case)
    
    passive_recirc_fraction = gnx.get_passive_recirc_fraction(param_df, case)
    passive_recirc_power = -ts_capacity * η * passive_recirc_fraction
    
    raw_fusion_core_power = fusion_core_power_in_zone(corepower_df, z)
    normalized_fusion_core_power = raw_fusion_core_power / ts_capacity
    
    raw_fusion_generation = fusion_generation_in_zone(power_df, z)
    norm_fusion_generation = raw_fusion_generation / gross_gen_cap
    
    raw_fusion_recirc = fusion_recirculation_in_zone(recirc_df, z)
    norm_recirc = raw_fusion_recirc / np.max(raw_fusion_recirc)
      
    raw_net_fusion_generation = net_fusion_generation_in_zone(power_df, recirc_df, z)
    norm_net_fusion_generation = raw_net_fusion_generation / long_run_average_power
        
    data = {'Active fraction': active_fraction,
            'Norm. fusion core power': normalized_fusion_core_power,
            'Norm. gross power generation': norm_fusion_generation, 
            'Norm. recirc power': norm_recirc, 
            'Norm. net fusion generation': norm_net_fusion_generation,
           }
    
    return data


# In[16]:


@functools.cache
def core_capacity_factor(scen, case):
    core_caps = gnx.ts_capacity_df(scen, case).CorePowerCap
    weighted_usage = []
    for zone in range(len(gnx.ZONE_NAMES)):
        plant_data = data_for_reactor_in_zone(scen, case, zone)
        active_frac = plant_data['Active fraction']
        mean_usage = np.mean(plant_data['Norm. fusion core power']) / active_frac
        weighted_usage.append(core_caps[zone] * mean_usage)
        
    return np.sum(weighted_usage) / np.sum(core_caps)

@functools.cache
def plant_capacity_factor(scen, case):
    core_caps = gnx.ts_capacity_df(scen, case).CorePowerCap
    weighted_usage = []
    for zone in range(len(gnx.ZONE_NAMES)):
        plant_data = data_for_reactor_in_zone(scen, case, zone)
        mean_usage = np.mean(plant_data['Norm. net fusion generation'])
        weighted_usage.append(core_caps[zone] * mean_usage)
        
    return np.sum(weighted_usage) / np.sum(core_caps)


# In[17]:


def actual_capacityfactors(scen, cases):
    vom, anncost = tableaux_positions(scen, cases)
    capacityfactors  = [plant_capacity_factor(scen, c) for c in cases]
    return vom, capacityfactors


# In[18]:


def actual_corecapacityfactors(scen, cases):
    vom, anncost = tableaux_positions(scen, cases)
    capacityfactors  = [core_capacity_factor(scen, c) for c in cases]
    return vom, capacityfactors


# In[19]:


def effective_COE(scen, cases):
    vom, anncost = tableaux_positions(scen, cases)
    capacityfactors  = [plant_capacity_factor(scen, c) for c in cases]
    interp_vom_to_cf = interp1d(vom, capacityfactors)
    return vom, 1000*anncost/(8760*interp_vom_to_cf(vom)) + np.array(vom)


# In[20]:


x, y = actual_capacityfactors("MedOppMustRun", [20400, 20401])


# In[22]:


chart_name = "Figure_7"

fig, ax = plt.subplots(2,3, figsize=(5.5,3), sharex=True, sharey="row")

case_ixs = [cases_10_gw, cases_50_gw, cases_100_gw]
case_labels = ["10 GW", "50 GW", "100 GW"]
linestyles = ('solid', 'solid', 'solid')
colors=["#AAAAAA", "#666666", "#333333"]
linewidths = [0.8, 1, 1.2]

data = dict()

for i, scen in enumerate(("LowOpp", "MedOpp", "HighOpp")):
    for cases, label, ls, c, lw in zip(case_ixs, case_labels, linestyles, colors, linewidths):
        kwargs = {'label': label, 'ls': ls, 'color': c, 'lw': lw}
        
        x, y = actual_capacityfactors(scen, cases)        
        data[f"VOM cost"] = x
        ax[0, i].plot(x, y, **kwargs)
        data[f"{scen} {label} capacity factor"] = y
        
        x, y = effective_COE(scen, cases)
        ax[1, i].plot(x, y,  **kwargs)
        data[f"{scen} {label} coe"] = y

        [ax[0, i].spines[side].set_visible(False) for side in ["top", "right"]]
        [ax[1, i].spines[side].set_visible(False) for side in ["top", "right"]]

# makes a horizontal line
# ax[-1,1].plot(*effective_COE("MedOppMustRun", cases_200_gw_mustrun), color='red')
        
for axx in ax[0,:]:
    axx.set_ylim(0.4,1)
    axx.set_xlim(0,45)
        
x = np.arange(0,50,1)
slope = 0.5
for axx in ax[1,:]:
    axx.set_ylim(40,120)
    axx.set_xlim(0,45)
    for y0 in [40, 60,80,100,120]:
        axx.plot(x, y0 + slope*x, color='gray', lw=0.2, ls='dashed')
        

#ax.set(xlabel="VOM cost / ($/MWh)", ylabel=)
fig.supxlabel("VOM cost / ($/MWh)", fontsize=8)
ax[0,0].set_ylabel("Plant\ncapacity factor", fontsize=8)
ax[1,0].set_ylabel("LCOE / ($/MWh)", fontsize=8)
#ax.set_xticklabels(ax.get_xticklabels(), fontsize=14)
ax[0,0].set_title("Low market opportunity", fontsize=8)
ax[0,1].set_title("Med opp.", fontsize=8)
ax[0,2].set_title("High opp.", fontsize=8)
ax[0,2].legend(loc=(0.38,0.52))

plt.savefig(gnx.product_path(chart_name+".pdf"))

plt.show()


# In[23]:


df = pd.DataFrame(data)
a = df.to_numpy()
columns = ", ".join(df.keys())

header = f"""{chart_name}:
Capacity factors and cost-of-electricity.

""" + gnx.generic_data_header + \
f"""

These are for mid-range-like plants.

The VOM cost and the cost of electricity are both in $/MWh.
Capacity factor measures the plant's annual output divided by its maximum possible output.
""" + \
columns

np.savetxt(gnx.product_path(chart_name+".csv"), a, fmt="%.3f", header=header, delimiter=',')


# # What about cases with storage?

# In[28]:


chart_name = "capacityfactor_and_coe"

fig, ax = plt.subplots(2,3, figsize=(5.5,3), sharex=True, sharey="row")

case_ixs = [cases_50_gw, cases_50_gw_stor, cases_100_gw, cases_100_gw_stor]
case_labels = ["50 GW", "50 GW + stor", "100 GW", "100 GW + stor"]
linestyles = ('solid', 'dashed', 'solid', 'dashed')
colors=["#666666", "#666666", "#333333", "#333333"]
linewidths = [1, 1, 1.2, 1.2]

data = dict()

for i, scen in enumerate(("LowOpp", "MedOpp", "HighOpp")):
    for cases, label, ls, c, lw in zip(case_ixs, case_labels, linestyles, colors, linewidths):
        kwargs = {'label': label, 'ls': ls, 'color': c, 'lw': lw}
        
        x, y = actual_capacityfactors(scen, cases)        
        data[f"VOM cost"] = x
        ax[0, i].plot(x, y, **kwargs)
        data[f"{scen} {label} capacity factor"] = y
        
        x, y = effective_COE(scen, cases)
        ax[1, i].plot(x, y,  **kwargs)
        data[f"{scen} {label} coe"] = y

        [ax[0, i].spines[side].set_visible(False) for side in ["top", "right"]]
        [ax[1, i].spines[side].set_visible(False) for side in ["top", "right"]]


for axx in ax[0,:]:
    axx.set_ylim(0.4,1)
    axx.set_xlim(0,45)
        
x = np.arange(0,50,1)
slope = 0.5
for axx in ax[1,:]:
    axx.set_ylim(40,120)
    axx.set_xlim(0,45)
    for y0 in [40, 60,80,100,120]:
        axx.plot(x, y0 + slope*x, color='gray', lw=0.2, ls='dashed')
        

#ax.set(xlabel="VOM cost / ($/MWh)", ylabel=)
fig.supxlabel("VOM cost / ($/MWh)", fontsize=8)
ax[0,0].set_ylabel("Plant\ncapacity factor", fontsize=8)
ax[1,0].set_ylabel("COE / ($/MWh)", fontsize=8)
#ax.set_xticklabels(ax.get_xticklabels(), fontsize=14)
ax[0,0].set_title("Low market opportunity", fontsize=8)
ax[0,1].set_title("Med opp.", fontsize=8)
ax[0,2].set_title("High opp.", fontsize=8)
ax[0,2].legend(loc=(0.38,0.52))

#plt.savefig(gnx.product_path(chart_name+".pdf"))

plt.show()


# In[ ]:




