# +
import genx_fusion_analysis_base as gnx

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
import matplotlib as mpl
# -

import calendar


# +
nice_gray = "#333333"

mpl.rcParams['font.family'] = 'sans-serif'
mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['axes.linewidth'] = 0.5
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.major.width'] = 0.5
mpl.rcParams['ytick.major.width'] = 0.5
mpl.rcParams['xtick.major.size'] = 2
mpl.rcParams['ytick.major.size'] = 2

mpl.rcParams['xtick.labelsize'] = 8
mpl.rcParams['ytick.labelsize'] = 8
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 10
# -

UTC_TO_EDT = -6


def shift_data(x):
    y = np.roll(x, UTC_TO_EDT)
    return y


scen = "MedOpp"
case_nost = 10004
case_stor = 10024

load = gnx.load_load_data_df(scen)

loadseries = shift_data(load.loc[:, load.columns.str.contains('Load_MW')].sum(axis=1).values[0:8759]/1e3)

# +
nost_pwr = gnx.power_hourly(scen, case_nost)
nost_chg = gnx.charge_hourly(scen, case_nost)

stor_pwr = gnx.power_hourly(scen, case_stor)
stor_chg = gnx.charge_hourly(scen, case_stor)
# -

nost_rec = gnx.fusion_recirc(scen, case_nost)
stor_rec = gnx.fusion_recirc(scen, case_stor)

nost_sta = gnx.load_fusion_starts_df(scen, case_nost)
stor_sta = gnx.load_fusion_starts_df(scen, case_stor)

stor_soc = gnx.ts_soc_df(scen, case_stor)

nost_cor = gnx.fusion_corepower_hourly(scen, case_nost)
stor_cor = gnx.fusion_corepower_hourly(scen, case_stor)


def get_fusion_starts(df):
    return shift_data(np.sum(df.loc[:, df.columns.str.contains('fusion')].values[2:, :], axis=1))


def get_tech_pwr(df: pd.DataFrame, tech: str, zone=None):
        is_tech = df.columns.str.contains(tech)
        is_interesting = is_tech
        if zone is not None:
            zone_name = gnx.ZONE_NAMES[zone] + '_'
            is_zone = df.columns.str.contains(zone_name)
            is_interesting = is_tech & is_zone
        return shift_data(df.loc[:, is_interesting].sum(axis=1).values/1000)


searchnames = {gnx.FUSION: 'fusion',
               gnx.NUCLEAR: 'nuclear',
               gnx.HYDROELECTRIC: 'hydroelectric_1',
               gnx.HYDROPUMPEDSTORAGE: 'pump',
               gnx.CCS: 'ccs',
               gnx.ZCF: 'avgcf',
               gnx.WIND: 'wind',
               gnx.SOLAR: 'solar',
              }

# +
a = 1000-36-4
b = a + 2*48 + 1

def nostpwr(tech):
    return get_tech_pwr(nost_pwr, tech)[a-1:b]
    
def nostchg(tech):
    return get_tech_pwr(nost_chg, tech)[a-1:b]

def nostrec():
    return get_tech_pwr(nost_rec, 'fusion')[a-1:b]

def nostcor():
    return get_tech_pwr(nost_cor, 'fusion')[a-1:b]


def storpwr(tech):
    return get_tech_pwr(stor_pwr, tech)[a-1:b]
    
def storchg(tech):
    return get_tech_pwr(stor_chg, tech)[a-1:b]

def storrec():
    return get_tech_pwr(stor_rec, 'fusion')[a-1:b]

def storcor():
    return get_tech_pwr(stor_cor, 'fusion')[a-1:b]

def storsoc():
    return get_tech_pwr(stor_soc, 'fusion')[a-1:b]


# +
pwr_labels = ['Fusion',
          'Fission',
          'Biomass',
          'Hydroelectric', 
          'NG-CCS',
          'ZCF CC & CT',
          'Wind',
          'Solar',
          gnx.DISTGEN,
          gnx.BATTERY,
          gnx.METALAIR,
          gnx.FLEX]

pwr_colors = [gnx.color_map[lab] for lab in pwr_labels]

sink_labels = [gnx.FUSION, gnx.HYDROELECTRIC, gnx.BATTERY, gnx.METALAIR, gnx.FLEX]
sink_colors = [gnx.color_map[lab] for lab in sink_labels]


# +
def main_stackplot_data(fpwr, fchg):
    return [fpwr('fusion'),
            fpwr('nuclear'),
            fpwr('biomass'),
            fpwr('hydroelectric_1')+fpwr('pump'),
            fpwr('ccs'),
            fpwr('avgcf'),
            fpwr('wind'),
            fpwr('solar')+fpwr('pv'),
            fpwr('distributed'),
            fpwr('battery'),
            fpwr('metal'),
            np.clip(fchg('mw') - fpwr('mw'), 0, None)]


def sink_stackplot_data(fpwr, fchg, frec):
    return [-frec(),
            -fchg('pump'),
            -fchg('battery'),
            -fchg('metal'),
            np.clip(fchg('mw')-fpwr('mw'), None, 0)]


# -

data_yn = main_stackplot_data(nostpwr, nostchg)
data_ys = main_stackplot_data(storpwr, storchg)

data_cn = sink_stackplot_data(nostpwr, nostchg, nostrec)
data_cs = sink_stackplot_data(storpwr, storchg, storrec)

data_sn = 2.2*get_fusion_starts(nost_sta)[a-1:b]
data_ss = 2.2*get_fusion_starts(stor_sta)[a-1:b]

product_name = "Figure_9"


def save_data(data, columns, figpart="a", msg=""):
    header = f"""{product_name} Part ({figpart}): Daily system cycles
    in the Medium market opportunity scenario with and without thermal storage.
    See cases 10004 and 100024
    """ + \
    gnx.generic_data_header + \
    f"""
    {msg}
    """ + \
    ",".join(columns)

    np.savetxt(gnx.product_path(f"{product_name}{figpart}.csv"), data, fmt="%.4f", header=header, delimiter=',')



# +
save_data(np.array(data_yn).T, columns=pwr_labels, figpart="a", msg="Electricity generation / GW")
save_data(np.array(data_ys).T, columns=pwr_labels, figpart="b", msg="Electricity generation / GW")

save_data(np.array(data_cn).T, columns=sink_labels, figpart="c", msg="Electricity sink / GW")
save_data(np.array(data_cs).T, columns=sink_labels, figpart="d", msg="Electricity sink / GW")

save_data(np.array([data_sn]).T, columns=["Fusion starts / GW_thermal"], figpart="e")
save_data(np.array([data_ss, storsoc()/1e3]).T, columns=["Fusion starts / GW_thermal", "Thermal storage / TWh_thermal"], figpart="f")

# +
import datetime

# written mostly by ChatGPT
def hour_to_datetime(hour: int) -> str:
    # Calculate the number of days
    days = hour // 24
    # Calculate the number of hours
    hour = hour % 24
    # Convert the number of days to a date
    date = datetime.datetime(2012, 1, 1) + datetime.timedelta(days=days, hours=hour)
    # Format the date and time as a string
    return date.strftime('%b %-d\n%-I') + date.strftime('%p').lower()


# +
fig, ax = plt.subplots(3,2, figsize=(7,4), sharex=True, sharey='row',
                       gridspec_kw={'height_ratios': [3, 1, 1],
                                     'wspace':0.12,
                                    'hspace': 0.35
                                    })

x = np.arange(a-1,b)

ax[0,0].stackplot(x, data_yn, labels=pwr_labels, colors=pwr_colors)
ax[1,0].stackplot(x, data_cn, labels=sink_labels, colors=sink_colors,)

stacks = ax[0,1].stackplot(x, data_ys, labels=pwr_labels, colors=pwr_colors,)
ax[1,1].stackplot(x, data_cs, labels=sink_labels, colors=sink_colors,)

ax[2,0].plot(x, data_sn, color='gray', lw=0.5)
ax[2,1].plot(x, data_ss, color='gray', lw=0.5)

twinx = ax[2,1].twinx()

twin_color = 'salmon'

twinx.plot(x, storsoc()/10, color=twin_color, lw=0.5)
twinx.spines['top'].set_visible(False)
twinx.spines['right'].set(color=twin_color)


twinx.set_ylabel('Fusion thermal\nstorage / TWh$_\mathrm{th}$', fontsize=8, color=twin_color)
twinx.set_ylim(0,250)
twinx.set_yticks([0,100,200])
twinx.set_yticklabels([0,1,2])

twinx.tick_params(axis='y', colors=twin_color)


for axx in ax[2,:]:
    axx.set_ylim(0,300)
    axx.spines['top'].set_visible(False)
    axx.spines['right'].set_visible(False)
     
ax[2,0].set_ylabel('Fusion core\nstarts / GW$_\mathrm{th}$', fontsize=8)

for axx in ax[0, :]:
    axx.plot(x, loadseries[a-1:b], 'k', lw=1, linestyle = '--', label = 'Load')

plt.xlim([a,b-1])

for axx in ax.flatten():
    axx.tick_params(axis='both', which='both', labelsize=8)

for axx in ax[0,:]:
    axx.set_ylim(0,1100)
    axx.set_yticks([0,200,400,600,800,1000])
    axx.spines['top'].set_visible(False)
    axx.spines['right'].set_visible(False)
    axx.tick_params(top=False, labeltop=False, bottom=True, labelbottom=True)

for axx in ax[1,:]:
    axx.set_ylim(-1100/3,0)
    axx.set_yticks([-200, 0], fontsize=8)
    axx.spines['bottom'].set_visible(False)
    axx.spines['right'].set_visible(False)
    axx.tick_params(top=True, labeltop=False, bottom=False, labelbottom=False)

tag_x = 0.05

def tag_top(ax, letter):
    ax.text(tag_x, 0.9, f'({letter})', va='bottom', ha='center', transform=ax.transAxes, size=8)

def tag_bottom(ax, letter):
    ax.text(tag_x, 0.035, f'({letter})', va='bottom', ha='center', transform=ax.transAxes, size=8)

tag_top(ax[0,0], 'a')
tag_top(ax[0,1], 'b')
tag_bottom(ax[1,0], 'c')
tag_bottom(ax[1,1], 'd')
tag_top(ax[2,0], 'e')
tag_top(ax[2,1], 'f')

hours_to_label = list(range(a, b, 12))

xlabels = [hour_to_datetime(hour) for hour in hours_to_label]

ax[0,0].set_xticks(hours_to_label, fontsize=6)
ax[0,0].set_xticklabels(xlabels, fontsize=6, color="#666666")
ax[0,1].set_xticks(hours_to_label, fontsize=6)
ax[0,1].set_xticklabels(xlabels, fontsize=6, color="#666666")



for axx in ax[-1, :]:
    axx.set_xticks(hours_to_label)
    axx.tick_params(top=False, labeltop=False, bottom=True, labelbottom=False)


ax[0,0].set_ylabel('Generation / GW', fontsize=8)
ax[1,0].set_ylabel('Sinks / GW', fontsize=8)

handles, labels = ax[0,1].get_legend_handles_labels()
ax[0,0].legend(handles[::-1],labels[::-1], bbox_to_anchor=(2.1, +1.45), fontsize=8, ncol=5)

plt.savefig(gnx.product_path('Figure_9.pdf'), bbox_inches='tight', dpi=300)
# -


