#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
import numpy as np
import pandas as pd

import calendar

import copy
import functools
import math
import os


# In[ ]:


my_scen = "MedOpp"
my_case = 0


# In[ ]:


def hourly_prices(scen, kase):
    """
    Returns
    -------
    df
    """
    df = gnx.load_price_data_df(scen, kase)
    return df[[str(i) for i in range(1, 21)]]


# In[ ]:


def price_data_for_zone(scen, kase, zone):
    df = gnx.load_price_data_df(scen, kase)
    return df[str(zone+1)].to_numpy()


# In[ ]:


prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']

nice_gray = "#333333"

mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.titlesize'] = 14


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = 14
mpl.rcParams['ytick.labelsize'] = 14
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 13


# In[ ]:


UTC_TO_EDT = -6


# In[ ]:


def highest_price(scen, kase):
    hp = hourly_prices(scen, kase)
    return np.max(hp.values)


# In[ ]:


downsampling_factor = 40


# In[ ]:


def plot_price_sorted_for_scenario(ax, scen, my_case=0):
    full_x = np.arange(1, 8761)
    sl = slice(downsampling_factor-1,None,downsampling_factor)
    
    for i in range(1, 20):
        hp = hourly_prices(scen, my_case)
        hp1 = hp[str(i)].to_numpy()
        sorted_prices = np.sort(hp1)
        top_zone = 12
        ax.plot(full_x[sl], sorted_prices[sl], lw=1)

    ax.set_ylim(0,250)
    ax.set_xlim(0, (60 + 8760))
    ax.grid()
    ax.set_xticks(np.linspace(0, 8760, 5, endpoint=True))
    ax.set_xticklabels(["0", "0.25", "0.5", "0.75", "1"])


# In[ ]:


def num_to_lowercase(n):
    return chr(ord("`") + 1 + n)


# In[ ]:


fig, axs = plt.subplots(5, 3, figsize=(12, 20), sharey=True, sharex=True)

plot_price_sorted_for_scenario(axs[1,0], "LowOpp")
plot_price_sorted_for_scenario(axs[1,1], "MedOpp")
plot_price_sorted_for_scenario(axs[1,2], "HighOpp")

plot_price_sorted_for_scenario(axs[0,0], "LowOppLowNuclear")
plot_price_sorted_for_scenario(axs[0,1], "MedOppLowNuclear")
plot_price_sorted_for_scenario(axs[0,2], "HighOppLowNuclear")

plot_price_sorted_for_scenario(axs[2,0], "LowOppNoNuc")
plot_price_sorted_for_scenario(axs[2,1], "MedOppNoNuc")
plot_price_sorted_for_scenario(axs[2,2], "HighOppNoNuc")


plot_price_sorted_for_scenario(axs[3,0], "LowOppNoNucOrCCS")
plot_price_sorted_for_scenario(axs[3,2], "HighOppNoNucOrCCS")


plot_price_sorted_for_scenario(axs[4,0], "CRLowOpp")
plot_price_sorted_for_scenario(axs[4,1], "CRMedOpp")
plot_price_sorted_for_scenario(axs[4,2], "CRHighOpp")



axs_size = np.size(axs)


[ax.spines["top"].set_visible(False) for ax in axs.reshape(axs_size)]
[ax.spines["right"].set_visible(False) for ax in axs.reshape(axs_size)]


secax = list(np.zeros_like(axs[:,2]))
for i, ax in enumerate(axs[:,2]):
    secax[i] = ax.secondary_yaxis('right', color=nice_gray)

[ax.spines["right"].set_visible(False) for ax in secax]
[ax.set_yticks([]) for ax in secax]

    
axbs = axs.reshape(axs_size)
for i, ax in enumerate(axbs):
    ax.text(0.85, 0.9, "("+num_to_lowercase(i)+")", ha='left', va='bottom', transform=ax.transAxes)
    

axs[3,1].set_visible(False)

axs[3,0].text(1.7, 0.5, "This scenario\n was not explored.", transform=axs[3,0].transAxes, ha='center')

# for ax in axs[2, 1:2]:
#     sc = ax.secondary_xaxis('bottom')
#     sc.set_xticks(np.linspace(0, 8760, 5, endpoint=True))
#     sc.set_xticklabels(["0", "25%", "50%", "75%", "100%"])
    
# axs[3,2].secondary_yaxis('left')


scen_group_labels = ["Low-cost nuclear", "Main Scenarios",  "No Fission", "No Fission or CCS", "Constrained renewables", ]
for i, l in enumerate(scen_group_labels):
    secax[i].set_ylabel(scen_group_labels[i], color=nice_gray, labelpad=35, rotation=270)
    
scen_group_labels = ["Low market opp.", "Medium market opp.", "High market opp."]
for i, ax in enumerate(axs[0,:]):
    ax.set_title(scen_group_labels[i], color=nice_gray)

    
fig.supxlabel("Fraction of hours in the year", y=0.08)
fig.supylabel("Locational marginal prices, $/MWh")

plt.savefig(gnx.product_path("SUPP_base_case_prices.pdf"))

plt.show()


# In[ ]:


prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']

mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.titlesize'] = 14
mpl.rcParams['xtick.labelsize'] = 14
mpl.rcParams['ytick.labelsize'] = 14
mpl.rcParams['legend.title_fontsize'] = 13


# In[ ]:


def set_monthlabels(axs):
    
    ax_last = axs[-1]
    
    month_ends = np.array([0,  31,  59,  90, 120, 151, 181, 212, 243, 273, 304, 334, 365])
    centers = month_ends[:-1] + (month_ends[1:] - month_ends[:-1]) / 2
    
    for ax in axs:
        ax.set_xticks(month_ends, minor=True)
        ax.tick_params(length=5, which='minor')
        ax.set_xticks(centers, minor=False)
        ax.tick_params(length=0, which='major')
        
    ax_last.set_xticklabels((calendar.month_abbr[i] for i in range(1, 13)))
 


# In[ ]:


def plot_prices_for_scenario(scen, my_case=0):
    fig, ax = plt.subplots(20, 1, figsize=(15,20), sharex=True)

    max_price = highest_price(scen, my_case)

    for i in range(20):
        a = price_data_for_zone(scen, my_case, i)
        a = np.roll(a, UTC_TO_EDT)
        data = a.reshape(365, 24).T / 200
        ax[i].pcolormesh(data, cmap=my_cmap, vmin=0, vmax=1)
        ax[i].set_ylabel(str(i+1))
        ax[i].set_yticks([0, 8, 16, 24])
        ax[i].set_yticklabels([0,8,16,24], size=8)
    
    plt.show()


# In[ ]:


fig = plt.figure(figsize=(11.5, 13.5))
subfigs = fig.subfigures(1, 2, wspace=0.00, width_ratios=[32,1])

axs = subfigs[0].subplots(14, 1, sharex=True)

axs_right = subfigs[1].subplots()


fig.tight_layout(h_pad=0.2)

my_cmap = mpl.cm.viridis.copy()
bounds = [0, 2, 5, 10, 20, 50, 100, 200]
norm = mpl.colors.BoundaryNorm(bounds, my_cmap.N, extend='both')

my_cmap.set_over('red')
my_cmap.set_under('blue')


common_opts = {'shading': 'flat', 'edgecolors': 'none', 'rasterized': True, 'snap': True}


def plot_prices_for_zone(ax, scen, lab):
    my_case = 0
    i = 11
    a = price_data_for_zone(scen, my_case, i)
    a = np.roll(a, UTC_TO_EDT)
    data = a.reshape(365, 24).T / 200
    cs = ax.pcolormesh(data, cmap=my_cmap, vmin=0, vmax=1, **common_opts)
    ax.set_ylabel(lab)
    ax.set_yticks([0, 8, 16, 24])
    ax.set_yticklabels([0,8,16,24], size=8)
    
scens = [
        "LowOppLowNuclear", "MedOppLowNuclear", "HighOppLowNuclear",
        "LowOpp", "MedOpp", "HighOpp",
         "LowOppNoNuc", "MedOppNoNuc", "HighOppNoNuc",
         "LowOppNoNucOrCCS", "HighOppNoNucOrCCS",
         "CRLowOpp", "CRMedOpp", "CRHighOpp",
]

labels = ["Low", "Med", "High",
          "Low", "Med", "High",
          "Low", "Med", "High",
          "Low", "High",
          "Low", "Med", "High", ]
caselabels = ["Low-cost fission", "Main scenarios", "No fission", "", "Constrained Renewables"]
axmap = [1, 4, 7, 9, 12]

for i, s in enumerate(scens):
    plot_prices_for_zone(axs[i], s, labels[i])

axs[-1].set_xlabel("Day of year")

labelx = -0.072

for i, lab in enumerate(caselabels):
    x = axs[axmap[i]]
    x.text(labelx, 0.45, lab, horizontalalignment='center', va='center', rotation=90, transform=x.transAxes)

line_x = -0.06
fig.add_artist(Line2D([line_x, line_x], [-1.35,2.25], transform=axs[1].transAxes, color='gray'))
fig.add_artist(Line2D([line_x, line_x], [-1.35,2.25], transform=axs[4].transAxes, color='gray'))
fig.add_artist(Line2D([line_x, line_x], [-1.35,2.25], transform=axs[7].transAxes, color='gray'))

fig.add_artist(Line2D([line_x, line_x], [-1.35,2.25], transform=axs[12].transAxes, color='gray'))

#last_ax = axs[-1]

fig.add_artist(Line2D([line_x, line_x], [-0.05,2.25], transform=axs[10].transAxes, color='gray'))
axs[10].text(labelx, 1.125, "No fission or CCS", horizontalalignment='center',
             va='center', rotation=90, transform=axs[10].transAxes)

#norm = mpl.colors.Normalize(vmin=0, vmax=200)
fig.colorbar(mpl.cm.ScalarMappable(norm=norm, cmap=my_cmap),
             cax=axs_right, orientation='vertical', label='Locational marginal prices, $/MWh', extend='both',
             spacing='proportional')

set_monthlabels(axs)

plt.savefig(gnx.product_path("SUPP_hourly_price_charts.pdf"), bbox_inches='tight', dpi=300)

plt.show()


# # Net-load duration curves

# In[ ]:


@functools.cache
def resource_category_dict(scen):
    """Lookup of category per resource
    
    Returns
    -------
    dict
    """
    all_resources = gnx.all_resources(scen)
    d = {g : gnx.generator_category(g, gnx.ZONE_NAMES) for g in all_resources}
    return d


# In[ ]:


def resources_in_zone(scen, z):
    zz = z + 1
    df = gnx.load_generators_data_raw(scen)
    return set(df[df.Zone == zz].Resource)


# In[ ]:


def resources_in_category(scen, category, in_zone=None):
    """Get all resources in a specific category
    Parameters
    ----------
    in_zone : None or int
        0-indexed zone number
        
    Returns
    -------
    set
    """
    d = resource_category_dict(scen)
    vals = {k for k, v in d.items() if v == category}
    
    if isinstance(in_zone, int):
        res_in_zone = resources_in_zone(scen, in_zone)
        vals = vals.intersection(res_in_zone)
    
    return vals


# In[ ]:


def wind_resources(scen, in_zone=None):
    """
    Parameters
    ----------
    scen : str
        Scenario folder
    in_zone : None or int
        0-indexed zone number
    """
    return resources_in_category(scen, gnx.WIND, in_zone)


# In[ ]:


def solar_resources(scen, in_zone=None):
    """
    Parameters
    ----------
    scen : str
        Scenario folder
    in_zone : None or int
        0-indexed zone number
    """
    return resources_in_category(scen, gnx.SOLAR, in_zone)


# In[ ]:


def storage_resources(scen, in_zone=None):
    """Resources that may have charging and power 
    """
    empty = set()
    hydro = resources_in_category(scen, gnx.HYDROELECTRIC, in_zone)
    batt = resources_in_category(scen, gnx.BATTERY, in_zone)
    metalair = resources_in_category(scen, gnx.METALAIR, in_zone)
    tot = empty.union(hydro, batt, metalair)
    return tot


# In[ ]:


def flex_resources(scen, in_zone=None):
    return resources_in_category(scen, gnx.FLEX, in_zone)


# In[ ]:


def variable_renewable_resources(scen, in_zone=None):
    """
    Parameters
    ----------
    scen : str
        Scenario folder
    in_zone : None or int
        0-indexed zone number
    """
    empty = set()
    wind = wind_resources(scen, in_zone)
    solar = solar_resources(scen, in_zone)
    dist_gen = resources_in_category(scen, gnx.DISTGEN, in_zone)
    
    return empty.union(wind, solar, dist_gen)


# In[ ]:


def storage_resource_charging(scen, kase, in_zone=None):
    resources = storage_resources(scen, in_zone)
    resources = list(resources)
    charge_df = gnx.charge_hourly(scen, kase)
    return np.sum(charge_df[resources].to_numpy(), axis=1)

def storage_resource_discharging(scen, kase, in_zone=None):
    resources = storage_resources(scen, in_zone)
    resources = list(resources)
    power_df = gnx.power_hourly(scen, kase)
    return np.sum(power_df[resources].to_numpy(), axis=1)


# In[ ]:


def flex_resource_charging(scen, kase, in_zone=None):
    resources = flex_resources(scen, in_zone)
    resources = list(resources)
    charge_df = gnx.charge_hourly(scen, kase)
    return np.sum(charge_df[resources].to_numpy(), axis=1)

def flex_resource_discharging(scen, kase, in_zone=None):
    resources = flex_resources(scen, in_zone)
    resources = list(resources)
    power_df = gnx.power_hourly(scen, kase)
    return np.sum(power_df[resources].to_numpy(), axis=1)


# In[ ]:


def variable_renewable_power_production(scen, kase, in_zone=None):
    """
    Parameters
    ----------
    scen : str
        Scenario folder
    kase : int
        Case
    in_zone : None or int
        0-indexed zone number
    """
    var_renew = variable_renewable_resources(scen, in_zone)
    var_renew = list(var_renew)
    power_df = gnx.power_hourly(scen, kase)
    return np.sum(power_df[var_renew].to_numpy(), axis=1)


# In[ ]:


def lines_serving_each_zone(network):
    """Get a dict of sets"""
    lines_serving = dict()
    num_lines = gnx.num_transmission_lines(network)
    for line in range(num_lines):
        z_start, z_end = gnx.line_endpoints(network, line)
        
        for z in (z_start, z_end):
            if not z in lines_serving:
                lines_serving[z] = {line}
            else:
                s = lines_serving[z]
                lines_serving[z] = s.union({line})
            
    return lines_serving


# In[ ]:


def lines_serving_a_zone(network, z):
    """
    Parameters
    ----------
    network: pd.DataFrame
    z : int
        Zero-indexed zones
    
    Returns
    -------
    Set of zero-indexed lines
    """
    lsez = lines_serving_each_zone(network)
    return lsez[z]


# In[ ]:


@functools.cache
def transmission_losses_associated_with_zone(scen, kase, z):
    network = gnx.load_original_network_connections(scen)
    lines_serving = lines_serving_a_zone(network, z)
    losses = np.zeros(8760)
    for line in lines_serving:
        losses += gnx.tlosses_hourly(scen, kase, line) / 2 / 1e3
       
    return losses.to_numpy()


# In[ ]:


@functools.cache
def net_load(scen, kase, zone):
    """Get the load minus any power provided by solar and wind
    
    Parameters
    ----------
    scen : str
        Scenario folder name
    kase : int
        case number
    zone : int
        Zone number, zero-indexed
        
    Returns
    -------
    array_like
    """
    zz = zone + 1
    raw_load = gnx.load_data_for_zone(zz, scen)   
    renew = variable_renewable_power_production(scen, kase, zone)   
    charging = storage_resource_charging(scen, kase, zone)
    discharging = storage_resource_discharging(scen, kase, zone)
    flex_charging = flex_resource_charging(scen, kase, zone)
    flex_discharging = flex_resource_discharging(scen, kase, zone)
    
    tlosses = transmission_losses_associated_with_zone(scen, kase, zone)
    
    recirc = gnx.fusion_recirc(scen, kase, zone)

    return raw_load - renew + charging - discharging - flex_charging + flex_discharging + tlosses + recirc


# In[ ]:


@functools.cache
def total_sorted_net_load(scen, kase):
    net_loads = np.array([net_load(scen, kase, i) for i in range(20)]).T
    total_net_loads = np.sum(net_loads, axis=1) / 1e6
    return np.sort(total_net_loads)


# In[ ]:


def plot_total_net_load_curve(ax, scen, kase, **kwargs):
    total_net_loads = total_sorted_net_load(scen, kase)
    ax.plot(total_net_loads, **kwargs)


# In[ ]:


def curve_intersection(level, net_load):
    return list(net_load > level).index(True), level


# In[ ]:


def meta_curve_intersection(scen, cases, cap_levels):
    intersections = []
    for i, c in enumerate(cases):
        total_net_load = total_sorted_net_load(scen, c)
        cap_level = cap_levels[i]
        i = curve_intersection(cap_level, total_net_load)
        intersections.append(i)
    return intersections


# In[ ]:


fig, axs = plt.subplots(3, 3, figsize=(12, 15), squeeze=False, sharex=True, sharey=True)
scen_low = "LowOpp"
scen_med = "MedOpp"
scen_high = "HighOpp"

penetration_labels = ["100 GW", "200 GW", "300 GW"]
opt_cases = [104, 106, 108]

midrange_cases = [10004, 10006, 10008]
all_midrange_cases = [10001, 10002, 10003, 10004, 10005, 10006, 10007, 10008, 10009]

all_opt_cases = [101, 102, 103, 104, 105, 106, 107, 108, 109]
all_pes_cases = range(1,10)
pes_cases = [4,6,8]

com_opts = {"clip_on": False}

fus_tw = [0.1, 0.2, 0.3]
all_fus_tw = [0.01, 0.02, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35]

def meta_intersection(ax, scen, cases, fus_cap, **kwargs):
    curveint = meta_curve_intersection(scen, cases, fus_cap)
    ax.plot(*np.array(curveint).T, color='black', lw=0.5, **kwargs)

def plot_multi(ax, scen, cases, all_cases, **kwargs):
    
    plot_total_net_load_curve(ax, scen, 0, label="0", color='gray', **kwargs)
    [plot_total_net_load_curve(ax, scen, c, label=penetration_labels[i], **kwargs) for i, c in enumerate(cases)]
    
    axs[0,0].hlines([0], 0, 8760, lw=0.5, color='gray')
    [ax.hlines(ftw, 0, 8760, color=gnx.colors[i], ls='dashed', lw=0.5) for i, ftw in enumerate(fus_tw)]
    
    meta_intersection(ax, scen, all_cases, all_fus_tw, **kwargs)

plot_multi(axs[0,0], scen_low, pes_cases, all_pes_cases, **com_opts)
plot_multi(axs[0,1], scen_med, pes_cases, all_pes_cases, **com_opts)
plot_multi(axs[0,2], scen_high, pes_cases, all_pes_cases, **com_opts)

plot_multi(axs[1,0], scen_low, midrange_cases, all_midrange_cases, **com_opts)
plot_multi(axs[1,1], scen_med, midrange_cases, all_midrange_cases, **com_opts)
plot_multi(axs[1,2], scen_high, midrange_cases, all_midrange_cases, **com_opts)

plot_multi(axs[2,0], scen_low, opt_cases, all_opt_cases, **com_opts)
plot_multi(axs[2,1], scen_med, opt_cases, all_opt_cases, **com_opts)
plot_multi(axs[2,2], scen_high, opt_cases, all_opt_cases, **com_opts)


scen_group_labels = ["Low market opp.", "Medium market opp.", "High market opp."]
for i, ax in enumerate(axs[0,:]):
    ax.set_title(scen_group_labels[i], color=nice_gray)

axs[1,1].legend(title="Fusion capacity", loc=(0.05, 0.52), ncol=1)

fig.supylabel("Net load / TW", x=0.04)
fig.supxlabel("Fraction of hours with net load below curve", y=0.08)

for ax in axs.reshape(9):
    [ax.spines[s].set_visible(False) for s in ["top", "right"]]
    ax.set_ylim(0, 0.8)
    ax.set_xlim(0, 8760)
    ax.set_xticks([0, 8760/2, 8760])
    ax.set_xticklabels(["0", "0.5", "1"])


secax = [0,0,0]
for i, ax in enumerate(axs[:,2]):
    secax[i] = ax.secondary_yaxis('right', color=nice_gray)
    
[ax.spines["right"].set_visible(False) for ax in secax]
[ax.set_yticks([]) for ax in secax]

axs[1,0].text(400, 0.4, "Intersection of\nnet load with\nfusion capacity", ha='left')

scen_group_labels = ["Pessmistic core", "Mid-range core", "Optimistic core"]
for i, l in enumerate(scen_group_labels):
    secax[i].set_ylabel(scen_group_labels[i], color=nice_gray, labelpad=35, rotation=270)

plt.savefig(gnx.product_path("SUPP_net_load_duration_curves_with_fusion.pdf"))    

plt.show()


# In[ ]:


fig, axs = plt.subplots(1, 5, figsize=(13, 5), sharey=True, sharex=True, squeeze=False)

com_opts = {"clip_on": False}

plot_total_net_load_curve(axs[0,1], "LowOpp", 0, label="Low", **com_opts)
plot_total_net_load_curve(axs[0,1], "MedOpp", 0, label="Medium", **com_opts)
plot_total_net_load_curve(axs[0,1], "HighOpp", 0, label="High", **com_opts)
axs[0,1].legend(title="Fusion market\nopportunity")

plot_total_net_load_curve(axs[0,0], "LowOppLowNuclear", 0, **com_opts)
plot_total_net_load_curve(axs[0,0], "MedOppLowNuclear", 0, **com_opts)
plot_total_net_load_curve(axs[0,0], "HighOppLowNuclear", 0, **com_opts)

plot_total_net_load_curve(axs[0,4], "CRLowOpp", 0, **com_opts)
plot_total_net_load_curve(axs[0,4], "CRMedOpp", 0, **com_opts)
plot_total_net_load_curve(axs[0,4], "CRHighOpp", 0, **com_opts)

plot_total_net_load_curve(axs[0,3], "LowOppNoNucOrCCS", 0, color=gnx.colors[0], **com_opts)
plot_total_net_load_curve(axs[0,3], "HighOppNoNucOrCCS", 0, color=gnx.colors[2], **com_opts)

plot_total_net_load_curve(axs[0,2], "LowOppNoNuc", 0, color=gnx.colors[0], **com_opts)
plot_total_net_load_curve(axs[0,2], "MedOppNoNuc", 0, color=gnx.colors[1], **com_opts)

plot_total_net_load_curve(axs[0,2], "HighOppNoNuc", 0, color=gnx.colors[2], **com_opts)

[ax.hlines([0], 0, 8760, color='black', lw=0.5) for ax in axs.reshape(5)]
    
scen_group_labels = ["Low-cost nuclear", "Main Scenarios", "No fission", "No Fission or CCS", "Constrained\nrenewables",]
for i, l in enumerate(scen_group_labels):
     axs[0,i].set_title(scen_group_labels[i], color=nice_gray)
    
for ax in axs.reshape(5):
    ax.set_xticks([0, 8760/2, 8760])
    ax.set_xticklabels(["0", "0.5", "1"])

    ax.set_xlim(0, 8760)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    #ax.spines["left"].set_visible(False)


    ax.set_clip_on(False)
    ax.set_ylim(0,0.6)

fig.supxlabel("Fraction of hours with net load below curve")
axs[0,0].set_ylabel("Net load / TW")

plt.savefig(gnx.product_path("SUPP_net_load_duration_curves_scenarios.pdf"))

plt.show()

