#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle

import genx_fusion_analysis_base as gnx
from genx_fusion_analysis_base import colors

from matplotlib.gridspec import GridSpec
import pandas as pd

import matplotlib as mpl


# Note: these cases have 6 hour ... 1 hour up/down times.

# In[ ]:


def mean_capacity_factor(scen, resource):
    gen_var = gnx.load_generators_variability(scen)
    return np.mean(gen_var[resource])


# In[ ]:


x_all = np.r_[10,20,50,100,150,200,250,300,350]
x_std = x_all[:-1]


# In[ ]:


a_std = np.arange(1,9)
a_all = np.arange(1, 10)


# In[ ]:


scen_low = "LowOpp"
scen_lowlownuc = "LowOppLowNuclear"
scen_med = "MedOpp"
scen_medlownuc = "MedOppLowNuclear"
scen_high = "HighOpp"
scen_highlownuc = "HighOppLowNuclear"
scen_lownonuc = "LowOppNoNucOrCCS"
scen_highnonuc = "HighOppNoNucOrCCS"
scen_crlow = "CRLowOpp"
scen_crmed = "CRMedOpp"
scen_crhigh = "CRHighOpp"



cases_nostor_pes = 0 + a_all

cases_lowstor_pes = 10 + a_all
cases_midstor_pes = 20 + a_all
cases_highstor_pes = 30 + a_all

cases_nostor_opt = 100 + a_all
cases_midstor_opt = 120 + a_all

cases_nostor_opt_stdgen = 200 + a_all

cases_midstor_opt_stdgen = 200 + 20 + np.r_[4]

cases_nostor_pes_advgen = 300 + a_all
cases_midstor_pes_advgen = 300 + 20 + a_all

cases_nostor_mid = 10000 + a_all
cases_midstor_mid = 10000 + 20 + a_all


# In[ ]:


# Supplementary Data: Nine scenarios


# In[ ]:





# In[ ]:


nice_gray = "#333333"

mpl.rcParams['font.size'] = 12
mpl.rcParams['axes.titlesize'] = 12


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 11


# In[ ]:


fusion_capacities = [0, 10, 20, 50, 100, 150, 200, 250, 300, 350]


# In[ ]:


def set_minorxticks(ax):
    ax.set_xticks(fusion_capacities, minor=True)
    ax.set_xlim((0,350))


def num_to_lowercase(n):
    return chr(ord("`") + 1 + n)


# In[ ]:


def three_cores_nostor(ax, scen, maxy=10):
    x, y_opt = max_plant_cost(scen, x_all, cases_nostor_opt)
    ax.plot(x, y_opt, color=colors[0])
    x, y_mid = max_plant_cost(scen, x_all, cases_nostor_mid)
    ax.plot(x, y_mid, color=colors[1], ls='dashed')
    x, y_pes = max_plant_cost(scen, x_all, cases_nostor_pes)
    ax.plot(x, y_pes, color=colors[2], ls='dotted')
    set_minorxticks(ax)
    ax.set_ylim(0, maxy)
    return {scen + ' Optimistic': y_opt, scen + ' Midrange': y_mid, scen + ' Pessimistic': y_pes}


# In[ ]:


fig, axs = plt.subplots(4, 3, figsize=(9,12), sharex=True, sharey=True)

styles = ['solid', 'dashed', 'dotted', 'dashed', 'dotted']

def max_plant_cost(scen, x, cases):
    duals = gnx.grab_duals(scen, cases)
    gen = gnx.grab_data("generation", scen, cases) / 1e3
    gen_per_cap = gen / x
    y = (duals + gnx.base_gen_cost * gen_per_cap)/1e3/0.078
    return x, y/1e3

generator_style = 'solid'

dump = {'Fusion capacity' :fusion_capacities[1:]}

dump = dump | three_cores_nostor(axs[1,0], scen_low)
dump = dump | three_cores_nostor(axs[1,1], scen_med)
dump = dump | three_cores_nostor(axs[1,2], scen_high)

dump = dump | three_cores_nostor(axs[0,0], scen_lowlownuc)
dump = dump | three_cores_nostor(axs[0,1], scen_medlownuc)
dump = dump | three_cores_nostor(axs[0,2], scen_highlownuc)

dump = dump | three_cores_nostor(axs[3,0], scen_crlow)
dump = dump | three_cores_nostor(axs[3,1], scen_crmed)
dump = dump | three_cores_nostor(axs[3,2], scen_crhigh)

dump = dump | three_cores_nostor(axs[2,0], scen_lownonuc)
dump = dump | three_cores_nostor(axs[2,2], scen_highnonuc)

legend_x = 1.25

secax = [0,0,0,0]

def my_capex_to_ann(x):
    return gnx.capex_to_ann(x * 1000)

def my_ann_to_capex(x):
    return gnx.ann_to_capex(x * 1000)

for i, ax in enumerate(axs[:,2]):
    secax[i] = ax.secondary_yaxis('right', functions=(my_capex_to_ann, my_ann_to_capex), color='gray')
    #secax.set_ylabel('Total annualized plant cost / (USD / kW·yr)', color='gray')

[ax.spines["top"].set_visible(False) for ax in axs.reshape(12)]
    
axbs = axs.reshape(12)
for i, ax in enumerate(axbs):
    ax.text(0.85, 0.9, "("+num_to_lowercase(i)+")", ha='left', va='bottom', transform=ax.transAxes)

#axs[3,0].set_visible(False)
#axs[3,0].text(0.85, 0.9, "(k)", ha='left', va='bottom', transform=axs[3,0].transAxes)
axs[2,1].set_visible(False)
#axs[3,2].text(0.85, 0.9, "(l)", ha='left', va='bottom', transform=axs[3,2].transAxes)

scen_group_labels = ["Low-cost nuclear", "Main Scenarios", "No Fission or CCS", "Constrained renewables", ]
for i, l in enumerate(scen_group_labels):
    secax[i].set_ylabel(scen_group_labels[i], color=nice_gray, labelpad=35, rotation=270)

scen_group_labels = ["Low market opp.", "Medium market opp.", "High market opp."]
for i, ax in enumerate(axs[0,:]):
    ax.set_title(scen_group_labels[i], color=nice_gray)

# for ax in axs[:,1]:
#     ax.set_yticklabels([])

# for ax in axs[:3,2]:
#     ax.set_yticklabels([])

[ax.secondary_xaxis('bottom') for ax in axs[2,:2]]
axs[2,2].secondary_yaxis('left')

fig.supylabel("Maximum plant capital cost, $/W", color=nice_gray)
fig.supxlabel("Fusion capacity / GW", color=nice_gray)
fig.text(0.945, 0.5, "Maximum total annual plant cost, $/kW·yr", rotation=90, color="gray", size=14, ha='center', va='center')

[ax.grid(color="#EEEEEE") for ax in axs.reshape(12)]

custom_lines = [Line2D([0], [0], color=colors[0], lw=2),
                Line2D([0], [0], color=colors[1], lw=2, ls='dashed'),
                Line2D([0], [0], color=colors[2], lw=2, ls='dotted')]

fig.legend(custom_lines,
                 ['Optimistic core', 'Mid-range core', 'Pessimistic core'],
                 loc=(0.37, 0.40), title="Fusion plants with")
fig.text(0.48, 0.33, 'This scenario\nwas not explored', ha='center', color=nice_gray, size=10)

plt.tight_layout()

product_name = "SUPP_capex_scenarios"

plt.savefig(gnx.product_path(product_name + '.pdf'), bbox_inches="tight")
plt.show()


# In[ ]:


df = pd.DataFrame(dump)

data = df.to_numpy()

columns = ", ".join(df.keys())

header = """Supplemental data:
Marginal value of the three reference fusion plants in eleven scenarios 
as a function of the system fusion capacity.
""" + gnx.generic_data_header + \
f"""

Capacity has units of GW
Other columns [scenario, reference reactor] have units of 2019$/W.

Note that to obtain the value in terms of 2019$/(kW·yr) (the right axis), multiply by {1e3*gnx.CAPEX_TO_ANN}.

""" + \
columns

np.savetxt(gnx.product_path(f"{product_name}.csv"), data, fmt="%.3f", header=header, delimiter=',')


# In[ ]:




