#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx


# In[ ]:


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle


import pandas as pd


# In[ ]:


def mean_capacity_factor(scen, resource):
    gen_var = gnx.load_generators_variability(scen)
    return np.mean(gen_var[resource])


# In[ ]:


x_all = np.r_[10,20,50,100,150,200,250,300,350]
x_std = x_all[:-1]


# In[ ]:


a_std = np.arange(1,9)
a_all = np.arange(1, 10)


# In[ ]:


#scen_low = "LowOpp"
#scen_lowlownuc = "LowOppLowNuclear"
scen_med = "MedOpp"

scen_medmustrun = "MedOppMustRun"

# scen_medlownuc = "MedOppLowNuclear"
# scen_high = "HighOpp"
# scen_highlownuc = "HighOppLowNuclear"
# scen_lownonuc = "LowOppNoNucOrCCS"
# scen_highnonuc = "HighOppNoNucOrCCS"



cases_nostor_pes = 0 + a_all

cases_lowstor_pes = 10 + a_all
cases_midstor_pes = 20 + a_all
cases_highstor_pes = 30 + a_all

cases_nostor_opt = 100 + a_all
cases_midstor_opt = 120 + a_all

cases_nostor_opt_stdgen = 200 + a_all

cases_midstor_opt_stdgen = 200 + 20 + np.r_[4]

cases_nostor_pes_advgen = 300 + a_all
cases_midstor_pes_advgen = 300 + 20 + a_all

cases_nostor_mid = 10000 + a_all
cases_midstor_mid = 10000 + 20 + a_all


# In[ ]:


def core_cost(scen, x, cases):
    duals = gnx.grab_duals(scen, cases)
    return x, (duals)/1e3/0.078

def max_plant_cost(scen, x, cases):
    duals = gnx.grab_duals(scen, cases)
    gen = gnx.grab_data("generation", scen, cases) / 1e3
    gen_per_cap = gen / x
    return x, (duals + gnx.base_gen_cost * gen_per_cap)/1e3/0.078

def generator_cost(scen, x, cases):
    gen = gnx.grab_data("generation", scen, cases) / 1e3
    gen_per_cap = gen / x
    return x, (gnx.base_gen_cost * gen_per_cap)/1e3/0.078


# In[ ]:


fig, ax = plt.subplots(1,2, figsize=(7,3.25), sharex=True, sharey=True)

styles = ['solid', 'dashed', 'dotted']
colors=gnx.colors

default_fontsize = 8

def value_of_flexibility(cases):
    x0, y0 = core_cost(scen_med, x_all, cases)
    yi, yi = core_cost(scen_medmustrun, x_all, cases)
    diff = y0-yi
    ix_of_100_GW = 3
    ix_of_50_GW = 2
    # 
    # print(diff[ix_of_50_GW])
    return x0, diff

ax[0].plot(*value_of_flexibility(cases_nostor_pes), color=colors[2], ls=styles[2], label='Pessimistic core.')
ax[0].plot(*value_of_flexibility(cases_nostor_mid), color=colors[1], ls=styles[1],  label='Mid-range core.')
ax[0].plot(*value_of_flexibility(cases_nostor_opt), color=colors[0], ls=styles[0], label='Optimistic core')

ax[1].plot(*value_of_flexibility(cases_midstor_opt), color=colors[0], ls=styles[0], label='Optimistic core, mid stor.')
ax[1].plot(*value_of_flexibility(cases_midstor_mid), color=colors[1], ls=styles[1],  label='Mid-range core, mid stor.')
ax[1].plot(*value_of_flexibility(cases_midstor_pes), color=colors[2], ls=styles[2], label='Pessimistic core, mid stor.')

legend_x = 1.25

[axx.set_xlim(0, 350) for axx in ax]
[axx.set_ylim(0, 800) for axx in ax]

cores = ['Optimistic core', 'Mid-range core', 'Pessimistic core'][::-1]

height = 0.74
legend_kwargs = {'fontsize':default_fontsize, 'title_fontsize': default_fontsize, 'framealpha':1}
#ax2.legend(title='Power conversion system\n costs for plants with:', loc=(legend_x, 0.0), **legend_kwargs)

fig.supxlabel("Fusion capacity / GW", fontsize=default_fontsize)
ax[0].set_ylabel("Marginal value of\nbeing dispatchable / (USD / kW)", fontsize=default_fontsize)
  
secax = ax[1].secondary_yaxis('right', functions=(gnx.capex_to_ann, gnx.ann_to_capex), color='gray')
secax.set_ylabel('As an annualized value\nfor marginal plant / (USD / kW·yr)', color='gray', fontsize=default_fontsize)
mileftax = ax[0].secondary_yaxis('right', functions=(gnx.capex_to_ann, gnx.ann_to_capex), color='gray')


leg = ax[0].legend(**legend_kwargs, loc='upper center')
leg.set_zorder(100)

ax[0].set_title("Without storage")
ax[1].set_title("With storage")

for axx, label in zip(ax, ("a", "b")):
    axx.text(0.05,0.9, f"({label})", transform=axx.transAxes)

plt.tight_layout()
plt.savefig(gnx.product_path('SUPP_marginal_value_dispatchability.pdf'), bbox_inches="tight")
plt.show()


# In[ ]:


fig, ax = plt.subplots(1,1, figsize=(11.5,5.5), sharex=True, sharey=True)

styles = ['solid', 'dashed', 'dotted']
colors=gnx.colors


ax.plot(*core_cost(scen_med, x_all, cases_midstor_opt), color=colors[1], ls=styles[1], label='Optimistic core, mid stor.')
ax.plot(*core_cost(scen_med, x_all, cases_midstor_mid), color=colors[1], ls=styles[1],  label='Mid-range core, mid stor.')
ax.plot(*core_cost(scen_med, x_all, cases_midstor_pes), color=colors[1], ls=styles[1], label='Pessimistic core, mid stor.')

ax.plot(*core_cost(scen_medmustrun, x_all, cases_midstor_opt), color=colors[1], ls=styles[0], label='Optimistic core, Must run, mid stor.')
ax.plot(*core_cost(scen_medmustrun, x_all, cases_midstor_mid), color=colors[1], ls=styles[0], label='Mid-range core, Must run, mid stor.')
ax.plot(*core_cost(scen_medmustrun, x_all, cases_midstor_pes), color=colors[1], ls=styles[0], label='Pessimistic core, Must run, mid stor.')

ax.plot(*core_cost(scen_med, x_all, cases_nostor_opt), color=colors[0], ls=styles[1], label='Optimistic core')
ax.plot(*core_cost(scen_med, x_all, cases_nostor_mid), color=colors[0], ls=styles[1],  label='Mid-range core.')
ax.plot(*core_cost(scen_med, x_all, cases_nostor_pes), color=colors[0], ls=styles[1], label='Pessimistic core.')

ax.plot(*core_cost(scen_medmustrun, x_all, cases_nostor_opt), color=colors[0], ls=styles[0], label='Optimistic core, Must run.')
ax.plot(*core_cost(scen_medmustrun, x_all, cases_nostor_mid), color=colors[0], ls=styles[0], label='Mid-range core, Must run.')
ax.plot(*core_cost(scen_medmustrun, x_all, cases_nostor_pes), color=colors[0], ls=styles[0], label='Pessimistic core, Must run.')

legend_x = 1.25

ax.set_xlim(0,350)
maxy = 7000
ax.set_ylim(0, maxy)

cores = ['Optimistic core', 'Mid-range core', 'Pessimistic core'][::-1]
height = 0.74
legend_kwargs = {'fontsize':12, 'title_fontsize': 12, 'framealpha':1}

ax.set_xlabel("Fusion capacity / GW")
ax.set_ylabel("Capital cost threshold\nfor marginal core / (USD / kW)")
    
secax = ax.secondary_yaxis('right', functions=(gnx.capex_to_ann, gnx.ann_to_capex), color='gray')
secax.set_ylabel('Annualized cost threshold\nfor marginal plant / (USD / kW·yr)', color='gray')

leg = ax.legend(**legend_kwargs, loc=(legend_x, 0.25))
leg.set_zorder(100)

ax.set_title("Comparing benefits of flexibility\n with benefits of thermal storage")

plt.tight_layout()
plt.show()


# In[ ]:




