#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx


# In[ ]:


import calendar

import numpy as np
import matplotlib as mpl
import matplotlib.colors as mcolors

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import pandas as pd

from mpl_toolkits.axes_grid1 import make_axes_locatable


# In[ ]:


def num_to_lowercase(n):
    return chr(ord("`") + 1 + n)


# In[ ]:


prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']

#mpl.rcParams['font.size'] = 14


nice_gray = "#333333"

mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.titlesize'] = 14


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = 14
mpl.rcParams['ytick.labelsize'] = 14
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = 13


# In[ ]:


UTC_TO_EDT = -6


# In[ ]:


η = 0.4


# In[ ]:


med_scen_parameters_df = gnx.load_core_parameters_df("MedOpp")


# In[ ]:


gnx.grab_duals("MedOpp", [10003, 10023]) / gnx.CAPEX_TO_ANN / 1e3


# In[ ]:


def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=-1):
    if n == -1:
        n = cmap.N
    new_cmap = mcolors.LinearSegmentedColormap.from_list(
         'trunc({name},{a:.2f},{b:.2f})'.format(name=cmap.name, a=minval, b=maxval),
         cmap(np.linspace(minval, maxval, n)))
    return new_cmap


# In[ ]:


def set_monthlabels(axs):
    
    ax_last = axs[-1]
    
    month_ends = np.array([0,  31,  59,  90, 120, 151, 181, 212, 243, 273, 304, 334, 365])
    centers = month_ends[:-1] + (month_ends[1:] - month_ends[:-1]) / 2
    
    for ax in axs:
        ax.set_xticks(month_ends, minor=True)
        ax.tick_params(length=5, which='minor')
        ax.set_xticks(centers, minor=False)
        ax.tick_params(length=0, which='major')
        
    ax_last.set_xticklabels((calendar.month_abbr[i] for i in range(1, 13)))
    
    #ax.set_xlabel('Day of year')


# # Load profile for each zone

# In[ ]:


common_opts = {'shading': 'flat', 'edgecolors': 'none', 'rasterized': True, 'snap': True}


fig, ax = plt.subplots(21, 1, figsize=(15,23), sharex=True)

for i in range(1,21):
    z = i-1
    a = gnx.load_data_for_zone(z+1, scen="MedOpp")
    a = np.roll(a, UTC_TO_EDT)
    data = a.reshape(365, 24).T / np.max(a)
    ax[i].pcolormesh(data, cmap='viridis', vmin=0, vmax=1, **common_opts)
    ax[i].set_ylabel(str(z+1) + ": " + gnx.ZONE_NAMES[z], size=10, rotation=0, ha='right')
    ax[i].set_yticks([0, 8, 16, 24])
    ax[i].set_yticklabels([0,8,16,24], size=8)
    
set_monthlabels(ax)

norm = mpl.colors.Normalize(vmin=0, vmax=1)
c_m = mpl.cm.viridis.copy()
s_m = mpl.cm.ScalarMappable(cmap=c_m, norm=norm)
s_m.set_array([])
ax[0].set_visible(False)
fig.colorbar(s_m, ax=[ax[0]], location='top', label='Normalized zonal load', aspect=15, shrink=0.3, fraction=1.0, anchor=(0.5,0))

# plt.savefig(gnx.product_path("SUPP_load_profiles.pdf"), bbox_inches='tight', dpi=400)

