#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx


# In[ ]:


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
import matplotlib as mpl

import pandas as pd


# In[ ]:


colors = gnx.colors

nice_gray = "#333333"

print_text_size = 8

mpl.rcParams['font.size'] = print_text_size
mpl.rcParams['axes.titlesize'] = print_text_size

mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = print_text_size
mpl.rcParams['ytick.labelsize'] = print_text_size
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = print_text_size


# Note: these cases have 6 hour ... 1 hour up/down times.

# In[ ]:


x_all = np.r_[10,20,50,100,150,200,250,300,350]
x_std = x_all[:-1]


# In[ ]:


miles_to_km = 1.609344


# In[ ]:


a_std = np.arange(1,9)
a_all = np.arange(1, 10)


# In[ ]:


scen_med = "MedOpp"
scen_high = "HighOpp"

cases_nostor_mid = 10000 + a_all
cases_midstor_mid = 10000 + 20 + a_all

cases_nostor_mid_noxmission = 30000 + a_all
cases_nostor_mid_anyxmission = 40000 + a_all


# In[ ]:


def transmission_expansion(scen, kase):
    network_2020 = gnx.load_base2020_network_connections(scen)
    network_2050_nx = gnx.load_network_expansion(scen, kase)
    total_2020 = sum(network_2020['Line_Max_Flow_MW'])
    total_2050 = total_2020 + sum(network_2050_nx['New_Trans_Capacity']) # yes, 2020 is correct
    return miles_to_km * total_2050/1e3


# In[ ]:


def max_plant_cost(scen, x, cases):
    duals = gnx.grab_duals(scen, cases)
    gen = gnx.grab_data("generation", scen, cases) / 1e3
    gen_per_cap = gen / x
    return x, (duals + gnx.base_gen_cost * gen_per_cap)/1e3/0.078


linewidth = 1

styles = {'HighOpp': {'color': colors[0], 'lw':linewidth},
         'MedOpp': {'color': colors[1], 'lw':linewidth}, 
          'LowOpp': {'color': colors[2], 'lw':linewidth}}

colors=gnx.colors

fig, ax = plt.subplots(1,2, figsize=(6.5, 3))

linestyles = ['solid', 'dashed', 'dotted']

a = list(range(10001,10010))
a = [0] + a

x = [0, 10, 20, 50, 100, 150, 200, 250, 300, 350]


for scen in ("HighOpp", "MedOpp", "LowOpp"):
    ax[0].plot(x, [transmission_expansion(scen, k) for k in a], **styles[scen], ls='dashed')

    
total_start_xmission = miles_to_km*sum(gnx.load_base2020_network_connections("MedOpp")['Line_Max_Flow_MW']) / 1e3

ax[0].hlines(total_start_xmission, 0, 350, color=nice_gray, lw=linewidth)

for scen in ("MedOpp", "HighOpp"):
    ax[0].plot(x[1:],[transmission_expansion(scen, k) for k in cases_nostor_mid_anyxmission], color=styles[scen]["color"], ls='dotted')

axl = ax[0].twinx()
axl.plot(np.NaN, np.NaN, label="High", c=colors[0])
axl.plot(np.NaN, np.NaN, label="Medium", c=colors[1])
axl.plot(np.NaN, np.NaN, label="Low", c=colors[2])
axl.get_yaxis().set_visible(False)


axl.legend(title="Market opportunity\n scenario", fontsize=print_text_size)
ax[0].set_ylim(0,1000)
ax[0].set_xlabel("Fusion capacity / GW", fontsize=print_text_size)
ax[0].set_ylabel("Final total tranmission / TW·km", fontsize=print_text_size)



for scen in (scen_med, scen_high):

    ax[1].plot(*max_plant_cost(scen, x_all, cases_nostor_mid_noxmission), ls=linestyles[0], **styles[scen])
    ax[1].plot(*max_plant_cost(scen, x_all, cases_nostor_mid), ls=linestyles[1], **styles[scen])
    ax[1].plot(*max_plant_cost(scen, x_all, cases_nostor_mid_anyxmission), ls=linestyles[2], **styles[scen])

ax[1].set_xlim(0,350)
maxy = 8700
ax[1].set_ylim(0, maxy)

scenarios = ['No Transmission Expansion', 'Normal', 'Unlimited expansion']
ax2 = ax[1].twinx()
for ss, sty in enumerate(styles):
    ax2.plot(np.NaN, np.NaN, ls=linestyles[ss],
             label=scenarios[ss], c='gray')
ax2.get_yaxis().set_visible(False)
#height = 0.74
legend_kwargs = {'fontsize':8, 'title_fontsize': 8, 'framealpha':1}
ax2.legend(title='Policy', loc='lower center', **legend_kwargs)

ax[1].set_xlabel("Fusion capacity / GW", fontsize=print_text_size)
ax[1].set_ylabel(r"Maximum marginal plant cost / (USD / kW)", fontsize=print_text_size)

#leg = ax.legend(title='Core design', **legend_kwargs, loc=(0.328,height))
#leg.set_zorder(100)

secax = ax[1].secondary_yaxis('right', functions=(gnx.capex_to_ann, gnx.ann_to_capex), color='gray')
secax.set_ylabel('Annualized cost / (USD / kW·yr)', color='gray', fontsize=8)

ax[1].text(70, 7000, "High Market Opp", color=colors[0], fontsize=print_text_size)
ax[1].text(50, 4200, "Med Market Opp", color=colors[1], fontsize=print_text_size)

for axx, label in zip(ax, ("a", "b")):
    axx.text(0.01,0.97, f"({label})", va='top', transform=axx.transAxes)

plt.tight_layout()
plt.savefig(gnx.product_path('SUPP_threshold_costs_transmissionsensitivity.pdf'), bbox_inches="tight")
plt.show()

