#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx


# In[ ]:


from cartopy import feature as cfeature
import cartopy.crs as ccrs
import cartopy.io.shapereader as shpreader
import geopandas as gpd
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as mc
from matplotlib.patches import Patch

import colorsys

import pandas as pd


# In[ ]:


prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']

mpl.rcParams['font.size'] = 14

η=0.4


# In[ ]:


nice_gray = "#333333"

mpl.rcParams['font.size'] = 12
mpl.rcParams['axes.titlesize'] = 12


mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
plt.rcParams['legend.title_fontsize'] = 11


# In[ ]:


ipm_model_regions = gpd.read_file("Twenty-Zone_Model_regions.shp")


# ### Map with zone names

# In[ ]:


def lighten_color(color, amount=0.5):
    """
    Lightens the given color by multiplying (1-luminosity) by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


# In[ ]:


def darken_color(color, amount=0.1):
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], amount, c[2])


# In[ ]:


d3_category20 = ['#1f77b4',
 '#aec7e8',
 '#ff7f0e',
 '#ffbb78',
 '#2ca02c',
 '#98df8a',
 '#d62728',
 '#ff9896',
 '#9467bd',
 '#c5b0d5',
 '#8c564b',
 '#c49c94',
 '#e377c2',
 '#f7b6d2',
 '#7f7f7f',
 '#c7c7c7',
 '#bcbd22',
 '#dbdb8d',
 '#17becf',
 '#9edae5',] # from some organization called d3


# In[ ]:


centroids_df = gnx.load_centroids()

model_regions = centroids_df.model_region.to_numpy()

num_zones = len(centroids_df)
centroids = [gnx.centroid_for_zone(z, centroids_df) for z in range(1,num_zones+1)]


# In[ ]:


my_crs = ccrs.epsg(3310)

my_colors = [lighten_color(col, 0.5) for col in d3_category20]
legend_elements = [Patch(facecolor=c, edgecolor=None,
                         label=l) for l, c in zip(gnx.ZONE_NAMES, my_colors)]


fig = plt.figure(figsize=(16, 16))

ax = plt.axes(projection=ccrs.PlateCarree())


for col, geom in zip(my_colors, ipm_model_regions.geometry):
    ax.add_geometries([geom], crs = my_crs, facecolor=col,
                  zorder = -20)
    
for c, z, color, in zip(centroids, gnx.ZONE_NAMES, d3_category20):
    ax.text(*c, z, ha='center', va='center', color=darken_color(color, 0.25), size=14)

plt.xlim(-108.25,-65)
plt.ylim(24,49.7)

ax.legend(handles=legend_elements, loc=4, fontsize=11)


ax.add_feature(cfeature.STATES, linewidth=0.4, edgecolor='lightgray', zorder=-3)

plt.savefig(gnx.product_path("SUPP_zone_map.png"), dpi=300, bbox_inches='tight')

plt.show()

