#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import genx_fusion_analysis_base as gnx
import pandas as pd
import numpy as np


# In[ ]:


max_loads = [max(gnx.load_data_for_zone(zz, "MedOpp")) for zz in range(1, 21)]


# In[ ]:


low_gen_var_df = gnx.load_generators_variability("LowOpp")
med_gen_var_df = gnx.load_generators_variability("MedOpp")
high_gen_var_df = gnx.load_generators_variability("HighOpp")


# In[ ]:


low_gen_df = gnx.load_generators_data("LowOpp")

med_gen_df = gnx.load_generators_data("MedOpp")

high_gen_df = gnx.load_generators_data("HighOpp")


# In[ ]:


ldv_resources = [r for r in med_gen_df.Resource if 'ldv_mw' in r]


# In[ ]:


def get_flex_rows(df, key):
    df_flex = df[df.Resource.str.contains(key)]
    df_flex = df_flex.sort_values('Zone')
    return df_flex


# In[ ]:


ldv_resources = get_flex_rows(med_gen_df, 'ldv_mw').Resource

res_wh_resources = get_flex_rows(med_gen_df, 'res_wh').Resource


# In[ ]:


def ldv_maxes(gen_df):
    df = get_flex_rows(gen_df, 'ldv_mw')
    return  np.array([df.Existing_Cap_MW], dtype='float')


def res_wh_maxes(gen_df):
    df = get_flex_rows(gen_df, 'res_wh')
    return  np.array([df.Existing_Cap_MW], dtype='float')


# In[ ]:


def two_columns(gen_df, gen_var_df):
    ldv = ldv_maxes(gen_df)
    res = res_wh_maxes(gen_df)
    a = (gen_var_df[ldv_resources] * ldv).max().to_numpy()
    b = (gen_var_df[res_wh_resources] * res).max().to_numpy()
    return np.array([a, b])


# In[ ]:


a = two_columns(low_gen_df, low_gen_var_df)
b = two_columns(med_gen_df, med_gen_var_df)
c = two_columns(high_gen_df, high_gen_var_df)


# In[ ]:


total_array = np.vstack([a,b,c]).T


# In[ ]:


def pprint_row(a):
    els = [str(int(np.round(el))) for el in a]
    return " & ".join(els) + " \\\\"
    


# In[ ]:


[print("\\verb|" + gnx.ZONE_NAMES[i] + "| & " + str(max_loads[i]) + " & " + pprint_row(r)) for i, r in enumerate(total_array)]

