#!/usr/bin/env python
# coding: utf-8

# This shows that most of the transmisison expansion takes place in the second period, 2035--2050

# In[2]:


import genx_fusion_analysis_base as gnx
import matplotlib.pyplot as plt
import numpy as np
import matplotlib as mpl


# In[3]:


colors = gnx.colors

nice_gray = "#333333"

print_text_size = 8

mpl.rcParams['font.size'] = print_text_size
mpl.rcParams['axes.titlesize'] = print_text_size

mpl.rcParams['axes.edgecolor'] = nice_gray
mpl.rcParams['axes.titlecolor'] = nice_gray
mpl.rcParams['axes.labelcolor'] = nice_gray
mpl.rcParams['text.color'] = nice_gray
mpl.rcParams['xtick.color'] = nice_gray
mpl.rcParams['ytick.color'] = nice_gray
mpl.rcParams['xtick.labelsize'] = print_text_size
mpl.rcParams['ytick.labelsize'] = print_text_size
mpl.rcParams['xtick.labelcolor'] = nice_gray
mpl.rcParams['ytick.labelcolor'] = nice_gray
mpl.rcParams['legend.title_fontsize'] = print_text_size


# In[4]:


styles = {'HighOpp': {'color': nice_gray, 'label': "High"},
         'MedOpp': {'color': 'gray', 'label': "Medium"}, 
          'LowOpp': {'color': '#AAA', 'label': "Low"}}


# # Tranmission expansion history

# In[5]:


def transmission_expansion_history_data(scen, kase):
    network_2020 = gnx.load_base2020_network_connections(scen)
    network_2035_nx = gnx.load_network_expansion(scen, 0, parent=True)
    network_2050_nx = gnx.load_network_expansion(scen, kase)

    total_2020 = sum(network_2020['Line_Max_Flow_MW'])
    total_2035 = total_2020 + sum(network_2035_nx['New_Trans_Capacity'])
    total_2050 = total_2020 + sum(network_2050_nx['New_Trans_Capacity']) # yes, 2020 is correct

    return [2020,2035,2050], np.array([total_2020, total_2035, total_2050])/1e3 # in TW·miles


# In[6]:


fig, ax = plt.subplots(figsize=(3,3))
kase = 0

for scen in ("HighOpp", "MedOpp", "LowOpp"):
    ax.plot(*transmission_expansion_history_data(scen, kase), **styles[scen])

# kase = 109
# ax.plot(*transmission_expansion_history_data("HighOpp", kase), label="High Opp", ls='dashed')
# ax.plot(*transmission_expansion_history_data("MedOpp", kase), label="Med", ls='dashed')
# ax.plot(*transmission_expansion_history_data("LowOpp", kase), label="Low", ls='dashed')

ax.set_ylim(0,350)
ax.legend(title="Market opportunity\nscenario")
ax.set_ylabel("TẆ·miles")
ax.set_xticks([2020,2035,2050])
plt.show()

